"""SQL repositories for entries."""
import logging
import typing
from typing import Dict, List, Optional, Generic, TypeVar

import injector
import regex
import sqlalchemy as sa
from sqlalchemy import func as sa_func
from sqlalchemy import sql
from sqlalchemy.orm import sessionmaker
import ulid

from karp.foundation.value_objects import UniqueId
from karp.foundation.events import EventBus
from karp.lex.domain import errors
from karp.lex.application import repositories

# from karp.domain.errors import NonExistingField, RepositoryError
from karp.lex.domain.entities.entry import (  # EntryRepositorySettings,; EntryRepository,; create_entry_repository,
    Entry,
    EntryOp,
    EntryStatus,
)
from karp.db_infrastructure import db
from karp.lex_infrastructure.sql import sql_models
from karp.db_infrastructure.sql_repository import SqlRepository
from karp.db_infrastructure.sql_unit_of_work import SqlUnitOfWork

logger = logging.getLogger(__name__)

DUPLICATE_PATTERN = r"Duplicate entry '(.+)' for key '(\w+)'"
DUPLICATE_PROG = regex.compile(DUPLICATE_PATTERN)
NO_PROPERTY_PATTERN = regex.compile(r"has no property '(\w+)'")


class SqlEntryRepository(repositories.EntryRepository, SqlRepository):
    def __init__(
        self,
        history_model,
        # runtime_model,
        resource_config: Dict,
        # resource_id: str,
        # mapped_class: Any
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        repositories.EntryRepository.__init__(self)
        SqlRepository.__init__(self, session=session)
        self.history_model = history_model
        # self.runtime_model = runtime_model
        self.resource_config = resource_config
        # self.mapped_class = mapped_class
        # self.resource_id = resource_id

    @classmethod
    def from_dict(
        cls,
        name: str,
        resource_config: typing.Dict,
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        # try:
        table_name = name
        # except KeyError:
        #     raise ValueError("Missing 'table_name' in settings.")

        # history_model = db.get_table(table_name)
        # if history_model is None:
        #     history_model = create_history_entry_table(table_name)
        # history_model.create(bind=db.engine, checkfirst=True)

        logger.info({"table_name": table_name})
        history_model = sql_models.get_or_create_entry_history_model(table_name)

        # runtime_table = db.get_table(runtime_table_name)
        # if runtime_table is None:
        #     runtime_table = create_entry_runtime_table(
        #         runtime_table_name, history_model, settings["config"]
        #     )

        if session:
            history_model.__table__.create(  # type:ignore [attr-defined]
                bind=session.bind, checkfirst=True
            )
        # runtime_table.create(bind=db.engine, checkfirst=True)
        # runtime_model = sql_models.get_or_create_entry_runtime_model(
        #     table_name, history_model, resource_config
        # )
        # if session:
        #     runtime_model.__table__.create(  # type:ignore [attr-defined]
        #         bind=session.bind, checkfirst=True
        #     )  # type:ignore [attr-defined]
        #     for child_model in runtime_model.child_tables.values():
        #         child_model.__table__.create(bind=session.bind, checkfirst=True)
        return cls(
            history_model=history_model,
            # runtime_model=runtime_model,
            resource_config=resource_config,
            # resource_id=settings["resource_id"],
            session=session,
        )

    @classmethod
    def _create_repository_settings(
        cls, resource_id: str, resource_config: typing.Dict
    ) -> typing.Dict:
        return {
            "table_name": resource_id,
            "resource_id": resource_id,
        }

    def _save(self, entry: Entry):
        self._check_has_session()

        history_id = self._insert_history(entry)

        # if entry.discarded:
        #     self._session.execute(
        #         sql.delete(self.runtime_model).where(
        #             self.runtime_model.entry_id == entry.entry_id
        #         )
        #     )
        #     return

        # runtime_entry_raw = self._entry_to_runtime_dict(history_id, entry)
        # try:
        #     entry_by_entry_id = (
        #         self._session.query(self.runtime_model)
        #         .filter_by(entry_id=entry.entry_id)
        #         .first()
        #     )
        #     entry_by_entity_id = (
        #         self._session.query(self.runtime_model)
        #         .filter_by(entity_id=entry.entity_id)
        #         .first()
        #     )

        #     logger.error(
        #         {
        #             "entry_by_entry_id": entry_by_entry_id,
        #             "entry_by_entity_id": entry_by_entity_id,
        #             "entry": entry.dict(),
        #         }
        #     )
        #     if not entry_by_entry_id:
        #         if not entry_by_entity_id:
        #             logger.warning("adding")
        #             return self._session.add(self.runtime_model(**runtime_entry_raw))
        #         else:
        #             logger.warning("entity_id but no entry_id")
        #             entry_by_entity_id.discarded = True
        #             return self._session.add(self.runtime_model(**runtime_entry_raw))
        #             return
        #     else:
        #         if not entry_by_entity_id:
        #             logger.warning("entry_id but no entity_id")
        #             for key, value in runtime_entry_raw.items():
        #                 setattr(entry_by_entry_id, key, value)
        #             return
        #         else:
        #             logger.warning("updating")
        #             for key, value in runtime_entry_raw.items():
        #                 setattr(entry_by_entry_id, key, value)
        #             return
        # if not entry_by_entity_id:
        #     # if entry discarded and replaced
        #     logger.error(
        #         {
        #             "entry_by_entry_id": entry_by_entry_id,
        #             "entry_by_entity_id": entry_by_entity_id,
        #         }
        #     )
        #     assert entry_by_entry_id.discarded
        #     raise RuntimeError(f"entry = {entry.dict()}")

        # if entry_by_entry_id.entity_id != entry.entity_id:
        #     logger.warn(
        #         "entity_id changed",
        #         extra={
        #             "entry_by_entry_id": entry_by_entry_id,
        #             "entry": entry.dict(),
        #         },
        #     )
        #     # raise RuntimeError(f'entry = {entry.dict()}')

        # if entry_by_entry_id.entry_id != entry.entry_id:
        #     logger.error(
        #         {
        #             "entry_by_entry_id": entry_by_entry_id,
        #             "entry_by_entity_id": entry_by_entity_id,
        #             "entry": entry.dict(),
        #         }
        #     )
        #     raise RuntimeError(f"entry = {entry.dict()}")
        # if entry_by_entity_id.entry_id != entry.entry_id:
        #     logger.error(
        #         {"entry_by_entry_id": entry_by_entry_id, "entry": entry.dict()}
        #     )
        #     raise RuntimeError(f"entry = {entry.dict()}")

        # for key, value in runtime_entry_raw.items():
        #     setattr(entry_by_entry_id, key, value)
        #            update_result = self._session.query(
        #                self.runtime_model
        #            ).filter_by(
        #                id=entry.id
        #            ).update(
        #                runtime_entry_raw,
        #            )
        #
        #            if update_result != 1:
        #                self._session.add(
        #                    self.runtime_model(
        #                        **runtime_entry_raw
        #                    )
        #                )
        # return self._session.add(runtime_entry)
        # except db.exc.DBAPIError as exc:
        #     logger.exception("db error")
        #     raise errors.RepositoryError("db failure") from exc

    def _update(self, entry: Entry):
        self._check_has_session()
        history_id = self._insert_history(entry)

        # current_db_entry = (
        #     self._session.query(self.runtime_model)
        #     .filter_by(entry_id=entry.entry_id)
        #     .first()
        # )

        # if not current_db_entry:
        #     raise errors.RepositoryError(f"Could not find {entry.entry_id}")

        # runtime_dict = self._entry_to_runtime_dict(history_id, entry)
        # for key, value in runtime_dict.items():
        #     setattr(current_db_entry, key, value)

        # @classmethod
        # def primary_key(cls):
        #     return "entry_id"

        # def move(self, entry: Entry, *, old_entry_id: str):
        #     self._check_has_session()

        #     db_entry = (
        #         self._session.query(self.runtime_model)
        #         .filter_by(entry_id=old_entry_id)
        #         .first()
        #     )
        #     if not db_entry:
        #         raise errors.RepositoryError(f"Could not find {entry.entry_id}")
        #     db_entry.discarded = True

        return self.save(entry)

    # def delete(self, entry: Entry):
    #     self._check_has_session()

    #     self._insert_history(entry)

    #     db_entry = (
    #         self._session.query(self.runtime_model)
    #         .filter_by(entry_id=entry.entry_id)
    #         .first()
    #     )
    #     if not db_entry:
    #         raise errors.RepositoryError(f"Could not find {entry.entry_id}")
    #     db_entry.discarded = True

    def _insert_history(self, entry: Entry):
        self._check_has_session()
        try:
            ins_stmt = db.insert(self.history_model)
            history_dict = self._entry_to_history_dict(entry)
            ins_stmt = ins_stmt.values(**history_dict)
            result = self._session.execute(ins_stmt)
            return result.lastrowid or result.returned_defaults["history_id"]
        except db.exc.DBAPIError as exc:
            raise errors.RepositoryError("db failure") from exc

    # def entry_ids(self) -> List[str]:
    #     self._check_has_session()
    #     query = self._session.query(self.runtime_model).filter_by(discarded=False)
    #     return [row.entry_id for row in query.all()]

    def entity_ids(self) -> List[str]:
        self._check_has_session()
        subq = (
            sql.select(
                self.history_model.entity_id,
                sa.func.max(self.history_model.last_modified).label("maxdate"),
            )
            .group_by(self.history_model.entity_id)
            .subquery("t2")
        )
        print(subq)
        stmt = sql.select(self.history_model).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            ),
        )
        stmt = stmt.order_by(self.history_model.last_modified.desc())
        query = self._session.execute(stmt).scalars()
        # query = self._session.query(self.history_model).filter_by(discarded=False)
        return [row.entity_id for row in query.all()]

    # def _by_entry_id(
    #     self,
    #     entry_id: str,
    # ) -> Optional[Entry]:
    #     self._check_has_session()
    #     subq = (
    #         sql.select(
    #             self.history_model.entity_id,
    #             sa.func.max(self.history_model.last_modified).label("maxdate"),
    #         )
    #         .group_by(self.history_model.entity_id)
    #         .subquery("t2")
    #     )

    #     stmt = sql.select(self.history_model).join(
    #         subq,
    #         sa.and_(
    #             self.history_model.entity_id == subq.c.entity_id,
    #             self.history_model.last_modified == subq.c.maxdate,
    #             self.history_model.entry_id == entry_id,
    #         ),
    #     )
    #     stmt = stmt.order_by(self.history_model.last_modified.desc())
    #     query = self._session.execute(stmt).scalars()
    #     row = query.first()
    #     return self._history_row_to_entry(row) if row else None

    def _by_id(
        self,
        id: str,
        *,
        version: Optional[int] = None,
        after_date: Optional[float] = None,
        before_date: Optional[float] = None,
        oldest_first: bool = False,
    ) -> Optional[Entry]:
        self._check_has_session()
        query = self._session.query(self.history_model)
        query = query.filter_by(entity_id=id)
        if version:
            query = query.filter_by(version=version)
        elif after_date is not None:
            query = query.filter(
                self.history_model.last_modified >= after_date
            ).order_by(self.history_model.last_modified)
        elif before_date:
            query = query.filter(
                self.history_model.last_modified <= before_date
            ).order_by(self.history_model.last_modified.desc())
        elif oldest_first:
            query = query.order_by(self.history_model.last_modified)
        else:
            query = query.order_by(self.history_model.last_modified.desc())
        row = query.first()
        return self._history_row_to_entry(row) if row else None

    # def history_by_entry_id(self, entry_id: str) -> List[Entry]:
    #     self._check_has_session()
    #     query = self._session.query(self.history_model)
    #     # query = query.join(
    #     #     self.runtime_table, self.history_model.c.id == self.runtime_table.c.id
    #     # )
    #     return query.filter_by(entry_id=entry_id).all()

    def teardown(self):
        """Use for testing purpose."""
        logger.info("starting teardown")
        #         for child_model in self.runtime_model.child_tables.values():
        #             print(f"droping child_model {child_model} ...")
        #             child_model.__table__.drop(bind=db.engine)
        #         print("droping runtime_model ...")
        #         self.runtime_model.__table__.drop(bind=db.engine)
        logger.info("droping history_model ...")
        self.history_model.__table__.drop(bind=self._session.bind)
        logger.info("dropped history_model")

        # db.metadata.drop_all(
        #     bind=db.engine, tables=[self.runtime_model, self.history_model]
        # )

    def all_entries(self) -> typing.Iterable[Entry]:
        self._check_has_session()

        subq = (
            sql.select(
                self.history_model.entity_id,
                sa.func.max(self.history_model.last_modified).label("maxdate"),
            )
            .group_by(self.history_model.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(self.history_model).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            ),
        )
        query = self._session.execute(stmt).scalars()
        # query = self._session.query(
        #     self.history_model).filter_by(discarded=False)
        return [self._history_row_to_entry(db_entry) for db_entry in query.all()]

    def get_total_entries(self) -> int:
        self._check_has_session()
        # query = self._session.query(self.runtime_model.discarded).filter_by(
        #     discarded=False
        # )
        # return query.count()
        return self.num_entities()

    def num_entities(self) -> int:
        self._check_has_session()
        # query = self._session.query(self.runtime_model.discarded).filter_by(
        #     discarded=False
        # )
        subq = (
            sql.select(
                self.history_model.entity_id,
                sa.func.max(self.history_model.last_modified).label("maxdate"),
            )
            .group_by(self.history_model.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(sa_func.count(self.history_model.entity_id)).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            ),
        )
        query = self._session.execute(stmt).scalar()
        return query

    def by_referenceable(self, filters: Optional[Dict] = None, **kwargs) -> List[Entry]:
        self._check_has_session()
        # query = self._session.query(self.runtime_model)
        # query = self._session.query(self.runtime_model, self.history_model).filter(
        #     self.runtime_model.history_id == self.history_model.history_id
        # )
        if filters is None:
            if not kwargs:
                raise ValueError("Must give 'filters' or kwargs")
            else:
                filters = kwargs

        subq = (
            sql.select(
                self.history_model.entity_id,
                sa.func.max(self.history_model.last_modified).label("maxdate"),
            )
            .group_by(self.history_model.entity_id)
            .subquery("t2")
        )
        query_and = sa.and_(
            self.history_model.entity_id == subq.c.entity_id,
            self.history_model.last_modified == subq.c.maxdate,
            self.history_model.discarded == False,
        )

        for filter_key, filter_value in filters.items():
            # query = query.filter(
            #     self.history_model.body.op("->")(filter_key) == filter_value
            # )
            query_and = query_and & (
                sa_func.json_extract(self.history_model.body, f"$.{filter_key}")
                == filter_value
            )
        stmt = sql.select(self.history_model).join(
            subq,
            query_and,
        )
        print(stmt)
        result = [
            self._history_row_to_entry(row)
            for row in self._session.execute(stmt).scalars().all()
        ]
        # joined_filters = []
        # simple_filters = {}

        # for filter_key in filters.keys():
        #     # tmp = collections.defaultdict(dict)
        #     if filter_key in self.resource_config[
        #         "referenceable"
        #     ] and self.resource_config["fields"][filter_key].get("collection"):
        #         logger.debug("collection field", extra={"filter_key": filter_key})
        #         # child_cls = self.runtime_model.child_tables[filter_key]
        #         # tmp[child_cls.__tablename__][filter_key] = filters[filter_key]
        #         # print(f"tmp.values() = {tmp.values()}")
        #         joined_filters.append({filter_key: filters[filter_key]})
        #         # query = query.filter(
        #         #     getattr(self.runtime_model, filter_key).any(filters[filter_key])
        #         # )
        #         # query = self._session.query(self.runtime_model.child_tables[filter_key])
        #         # return query.all()
        #     else:
        #         simple_filters[filter_key] = filters[filter_key]
        #     # joined_filters.extend(tmp.values())

        # try:
        #     query = query.filter_by(**simple_filters)
        # except db.exc.InvalidRequestError as exc:
        #     match = NO_PROPERTY_PATTERN.search(str(exc))
        #     if match:
        #         raise errors.NonExistingField(match.group(1)) from exc
        #     else:
        #         raise errors.RepositoryError("Unknown invalid request") from exc

        # for child_filters in joined_filters:
        #     child_table_name = list(child_filters.keys())[0]
        #     logger.debug("child table name", extra={"table_name": child_table_name})
        #     child_cls = self.runtime_model.child_tables[child_table_name]
        #     child_query = self._session.query(child_cls).filter_by(**child_filters)
        #     for child_e in child_query:
        #         logger.debug("child hit", extra={"child_hit": child_e})
        #     query = query.join(child_cls).filter_by(**child_filters)
        # # result = query.filter_by(**kwargs).all()
        # # # query = self._session.query(self.history_model)
        # # # query = query.join(
        # # #     self.runtime_table,
        # # #     self.history_model.c.history_id == self.runtime_table.c.history_id,
        # # # )
        # # # result = query.filter_by(larger_place=7).all()
        # # print(f"result = {result}")
        # # return result
        # # return query.all()
        # return [self._history_row_to_entry(db_entry) for _, db_entry in query.all()]

        return result

    def get_history(
        self,
        user_id: Optional[str] = None,
        entry_id: Optional[str] = None,
        from_date: Optional[float] = None,
        to_date: Optional[float] = None,
        from_version: Optional[int] = None,
        to_version: Optional[int] = None,
        offset: int = 0,
        limit: int = 100,
    ):
        self._check_has_session()
        query = self._session.query(self.history_model)
        if user_id:
            query = query.filter_by(last_modified_by=user_id)
        if entry_id:
            query = query.filter_by(entity_id=entry_id)
        if entry_id and from_version:
            query = query.filter(self.history_model.version >= from_version)
        elif from_date is not None:
            query = query.filter(self.history_model.last_modified >= from_date)
        if entry_id and to_version:
            query = query.filter(self.history_model.version < to_version)
        elif to_date is not None:
            query = query.filter(self.history_model.last_modified <= to_date)

        paged_query = query.limit(limit).offset(offset)
        total = query.count()
        return [self._history_row_to_entry(row) for row in paged_query.all()], total

    # def _entry_to_history_row(
    #     self, entry: Entry
    # ) -> Tuple[None, UUID, str, int, float, str, Dict, EntryStatus, str, EntryOp, bool]:
    #     return (
    #         None,  # history_id
    #         entry.id,  # id
    #         entry.entry_id,  # entry_id
    #         entry.version,  # version
    #         entry.last_modified,  # last_modified
    #         entry.last_modified_by,  # last_modified_by
    #         entry.body,  # body
    #         entry.status,  # version
    #         entry.message,  # message
    #         entry.op,  # op
    #         entry.discarded,
    #     )

    def _entry_to_history_dict(
        self, entry: Entry, history_id: Optional[int] = None
    ) -> Dict:
        return {
            "history_id": history_id,
            "entity_id": entry.entity_id,
            # "entry_id": str(entry.entity_id),
            "version": entry.version,
            "last_modified": entry.last_modified,
            "last_modified_by": entry.last_modified_by,
            "body": entry.body,
            "status": entry.status,
            "message": entry.message,
            "op": entry.op,
            "discarded": entry.discarded,
            "repo_id": entry.repo_id,
        }

    def _history_row_to_entry(self, row) -> Entry:
        print(f"{row.message=}")
        return Entry(
            body=row.body,
            message=row.message,
            status=row.status,
            op=row.op,
            entity_id=row.entity_id,
            last_modified=row.last_modified,
            last_modified_by=row.last_modified_by,
            discarded=row.discarded,
            version=row.version,
            repository_id=row.repo_id,
            # entry_id=row.entry_id,
            # resource_id=self.resource_id,
        )

    def _entry_to_runtime_dict(self, history_id: int, entry: Entry) -> Dict:
        _entry = {
            "entry_id": str(entry.entity_id),
            "history_id": history_id,
            "entity_id": entry.entity_id,
            "discarded": entry.discarded,
        }
        for field_name in self.resource_config.get("referenceable", ()):
            field_val = entry.body.get(field_name)
            if field_val is None:
                continue
            if self.resource_config["fields"][field_name].get("collection"):
                child_table = self.runtime_model.child_tables[field_name]
                for elem in field_val:
                    if field_name not in _entry:
                        _entry[field_name] = []
                    _entry[field_name].append(child_table(**{field_name: elem}))
            else:
                _entry[field_name] = field_val
        return _entry


class SqlEntryUnitOfWork(
    SqlUnitOfWork,
    repositories.EntryUnitOfWork,
):
    repository_type: str = "sql_entries_base"

    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
        **kwargs,
    ):
        SqlUnitOfWork.__init__(self)
        repositories.EntryUnitOfWork.__init__(self, event_bus=event_bus, **kwargs)
        self.session_factory = session_factory
        self._entries = None
        self._session = None

    def _begin(self):
        if self._session is None:
            self._session = self.session_factory()
        if self._entries is None:
            self._entries = SqlEntryRepository.from_dict(
                name=self.table_name(),
                resource_config=self.config,
                session=self._session,
            )
        return self

    def table_name(self) -> str:
        return self.name

    @property
    def repo(self) -> SqlEntryRepository:
        if self._entries is None:
            raise RuntimeError("No entries")
        return self._entries

    @classmethod
    def from_dict(cls, settings: typing.Dict, resource_config, **kwargs):
        return cls(repo_settings=settings, resource_config=resource_config, **kwargs)

    def collect_new_events(self) -> typing.Iterable:
        if self._entries:
            return super().collect_new_events()
        else:
            return []


class SqlEntryUnitOfWorkV1(SqlEntryUnitOfWork):
    repository_type: str = "sql_entries_v1"


class SqlEntryUnitOfWorkV2(SqlEntryUnitOfWork):
    repository_type: str = "sql_entries_v2"

    def table_name(self) -> str:
        u = ulid.from_uuid(self.entity_id)
        random_part = u.randomness().str
        return f"{self.name}_{random_part}"


# ===== Value objects =====
# class SqlEntryRepositorySettings(EntryRepositorySettings):
#     def __init__(self, *, table_name: str, config: Dict):
#         self.table_name = table_name
#         self.config = config


# @create_entry_repository.register(SqlEntryRepositorySettings)
# def _(settings: SqlEntryRepositorySettings) -> SqlEntryRepository:
#     history_model = sql_models.get_or_create_entry_history_model(settings.table_name)

#     runtime_table_name = f"runtime_{settings.table_name}"

#     runtime_model = sql_models.get_or_create_entry_runtime_model(
#         runtime_table_name, history_model, settings.config
#     )
#     return SqlEntryRepository(history_model, runtime_model, settings.config)
SqlEntryUowType = TypeVar("SqlEntryUowType", bound=SqlEntryUnitOfWork)


class SqlEntryUowCreator(Generic[SqlEntryUowType]):
    repository_type: str = "repository_type"

    @injector.inject
    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
    ):
        self._session_factory = session_factory
        self.event_bus = event_bus
        self.cache = {}

    def _create_uow(self, **kwargs) -> SqlEntryUowType:
        raise NotImplementedError(f"please implement this for {self}")

    def __call__(
        self,
        entity_id: UniqueId,
        name: str,
        config: Dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> SqlEntryUowType:
        if entity_id not in self.cache:
            self.cache[entity_id] = self._create_uow(
                entity_id=entity_id,
                name=name,
                config=config,
                connection_str=connection_str,
                last_modified_by=user,
                message=message,
                last_modified=timestamp,
                session_factory=self._session_factory,
                event_bus=self.event_bus,
            )
        return self.cache[entity_id]


class SqlEntryUowV1Creator(SqlEntryUowCreator[SqlEntryUnitOfWorkV1]):
    repository_type: str = "sql_entries_v1"

    def _create_uow(self, **kwargs) -> SqlEntryUnitOfWorkV1:
        return SqlEntryUnitOfWorkV1(**kwargs)


class SqlEntryUowV2Creator(SqlEntryUowCreator[SqlEntryUnitOfWorkV2]):
    repository_type: str = "sql_entries_v2"

    def _create_uow(self, **kwargs) -> SqlEntryUnitOfWorkV2:
        return SqlEntryUnitOfWorkV2(**kwargs)



from .entry_handlers import (
    AddingEntry,
    AddingEntries,
    DeletingEntry,
    ImportingEntries,
    UpdatingEntry,
)
from .entry_repo_handlers import CreatingEntryRepo, DeletingEntryRepository
from .resource_handlers import (
    CreatingResource,
    DeletingResource,
    PublishingResource,
    SettingEntryRepoId,
    UpdatingResource,
)


__all__ = [
    "AddingEntries",
    "AddingEntry",
    "ImportingEntries",
]



import json
import logging
import re
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import elasticsearch
import elasticsearch.helpers  # pyre-ignore
import elasticsearch_dsl as es_dsl  # pyre-ignore
from tatsu import exceptions as tatsu_exc

from karp import search

# from karp import query_dsl
from karp.search.application.queries import (
    QueryRequest,
)
from karp.search.domain import errors
from karp.search.domain.errors import (
    UnsupportedField,
)  # IncompleteQuery,; UnsupportedQuery,
from karp.lex.domain.entities.entry import Entry
from karp.lex.domain.entities.resource import Resource
from karp.search.domain import query_dsl
from karp.search_infrastructure.elasticsearch6 import es_config
from .es_query import EsQuery

# from karp.query_dsl import basic_ast as ast, op, is_a


logger = logging.getLogger(__name__)

KARP_CONFIGINDEX = "karp_config"
KARP_CONFIGINDEX_TYPE = "configs"


class EsQueryBuilder(query_dsl.NodeWalker):
    def walk_object(self, node):
        return node

    def walk__and(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result & self.walk(n)

        return result

    def walk__contains(self, node):
        return es_dsl.Q(
            "regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}.*"}
        )

    def walk__endswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}"})

    # def walk__equals(self, node):
    #     return es_dsl.Q(
    #         "match",
    #         **{
    #             self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}
    #         },
    #     )

    def walk__equals(self, node):
        return es_dsl.Q(
            "match",
            **{
                self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}
            },
        )

    def walk__exists(self, node):
        return es_dsl.Q("exists", field=self.walk(node.field))

    def walk__freergxp(self, node):
        return es_dsl.Q(
            "query_string", query=f"/{self.walk(node.arg)}/", default_field="*"
        )

    def walk__freetext_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg), fuzziness=1)

    def walk__freetext_any_but_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg))

    def walk_range(self, node):
        return es_dsl.Q(
            "range",
            **{self.walk(node.field): {self.walk(node.op): self.walk(node.arg)}},
        )

    walk__gt = walk_range
    walk__gte = walk_range
    walk__lt = walk_range
    walk__lte = walk_range

    def walk__missing(self, node):
        return es_dsl.Q(
            "bool", must_not=es_dsl.Q("exists", field=self.walk(node.field))
        )

    def walk__not(self, node):
        must_nots = [self.walk(expr) for expr in node.exps]
        return es_dsl.Q("bool", must_not=must_nots)

    def walk__or(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result | self.walk(n)

        return result

    def walk__regexp(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): self.walk(node.arg)})

    def walk__startswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f"{self.walk(node.arg)}.*"})


class Es6SearchService(search.SearchService):
    def __init__(self, es: elasticsearch.Elasticsearch):
        self.es: elasticsearch.Elasticsearch = es
        self.query_builder = EsQueryBuilder()
        self.parser = query_dsl.KarpQueryV6Parser(
            semantics=query_dsl.KarpQueryV6ModelBuilderSemantics()
        )
        if not self.es.indices.exists(index=KARP_CONFIGINDEX):
            self.es.indices.create(
                index=KARP_CONFIGINDEX,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1,
                        "refresh_interval": -1,
                    },
                    "mappings": {
                        KARP_CONFIGINDEX_TYPE: {
                            "dynamic": False,
                            "properties": {"index_name": {"type": "text"}},
                        }
                    },
                },
            )
        analyzed_fields, sortable_fields = self._init_field_mapping()
        self.analyzed_fields: Dict[str, List[str]] = analyzed_fields
        self.sortable_fields: Dict[str, Dict[str, List[str]]] = sortable_fields

    def _get_index_name_for_resource(self, resource_id: str) -> str:
        res = self.es.get(
            index=KARP_CONFIGINDEX, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
        )
        return res["_source"]["index_name"]

    @staticmethod
    def get_analyzed_fields_from_mapping(
        properties: Dict[str, Dict[str, Dict[str, Any]]]
    ) -> List[str]:
        analyzed_fields = []

        for prop_name, prop_values in properties.items():
            if "properties" in prop_values:
                res = Es6SearchService.get_analyzed_fields_from_mapping(
                    prop_values["properties"]
                )
                analyzed_fields.extend([prop_name + "." + prop for prop in res])
            else:
                if prop_values["type"] == "text":
                    analyzed_fields.append(prop_name)
        return analyzed_fields

    def _init_field_mapping(
        self,
    ) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
        """
        Create a field mapping based on the mappings of elasticsearch
        currently the only information we need is if a field is analyzed (i.e. text)
        or not.
        """

        field_mapping: Dict[str, List[str]] = {}
        sortable_fields = {}
        # Doesn't work for tests, can't find resource_definition
        # for resource in resourcemgr.get_available_resources():
        #     mapping = self.es.indices.get_mapping(index=resource.resource_id)
        #     field_mapping[resource.resource_id] = parse_mapping(
        #         next(iter(mapping.values()))['mappings']['entry']['properties']
        #     )
        aliases = self._get_all_aliases()
        mapping: Dict[
            str, Dict[str, Dict[str, Dict[str, Dict]]]
        ] = self.es.indices.get_mapping()
        # print(f"mapping = {mapping}")
        for (alias, index) in aliases:
            if (
                "mappings" in mapping[index]
                and "entry" in mapping[index]["mappings"]
                and "properties" in mapping[index]["mappings"]["entry"]
            ):
                field_mapping[
                    alias
                ] = Es6SearchService.get_analyzed_fields_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
                sortable_fields[
                    alias
                ] = Es6SearchService.create_sortable_map_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
        return field_mapping, sortable_fields

    def _get_index_mappings(
        self, index: Optional[str] = None
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Dict]]]]:
        kwargs = {"index": index} if index is not None else {}
        return self.es.indices.get_mapping(**kwargs)

    def _get_all_aliases(self) -> List[Tuple[str, str]]:
        """
        :return: a list of tuples (alias_name, index_name)
        """
        result = self.es.cat.aliases(h="alias,index")
        logger.debug("ES aliases and indicies", extra={"result": result})
        index_names = []
        for index_name in result.split("\n")[:-1]:
            logger.debug("existing index", extra={"index_name": index_name})
            if index_name[0] != ".":
                opt_match = re.search(r"([^ ]*) +(.*)", index_name)
                if opt_match:
                    groups = opt_match.groups()
                    alias = groups[0]
                    index = groups[1]
                    index_names.append((alias, index))
        return index_names

    def build_query(self, args, resource_str: str) -> EsQuery:
        query = EsQuery()
        query.parse_arguments(args, resource_str)
        return query

    def _format_result(self, resource_ids, response):
        logger.debug("_format_result called", extra={"resource_ids": resource_ids})

        def format_entry(entry):

            dict_entry = entry.to_dict()
            version = dict_entry.pop("_entry_version", None)
            last_modified_by = dict_entry.pop("_last_modified_by", None)
            last_modified = dict_entry.pop("_last_modified", None)
            return {
                "id": entry.meta.id,
                "version": version,
                "last_modified": last_modified,
                "last_modified_by": last_modified_by,
                "resource": next(
                    resource
                    for resource in resource_ids
                    if entry.meta.index.startswith(resource)
                ),
                "entry": dict_entry,
            }

        result = {
            "total": response.hits.total,
            "hits": [format_entry(entry) for entry in response],
        }
        return result

    def query(self, request: QueryRequest):
        logger.info("query called", extra={"request": request})
        query = EsQuery.from_query_request(request)
        return self.search_with_query(query)

    def query_split(self, request: QueryRequest):
        logger.info("query_split called", extra={"request": request})
        query = EsQuery.from_query_request(request)
        query.split_results = True
        return self.search_with_query(query)

    def search_with_query(self, query: EsQuery):
        logger.info("search_with_query called", extra={"query": query})
        es_query = None
        if query.q:
            try:
                model = self.parser.parse(query.q)
                es_query = self.query_builder.walk(model)
            except tatsu_exc.FailedParse as err:
                logger.debug("Parse error", extra={"err": err})
                raise errors.IncompleteQuery(
                    failing_query=query.q, error_description=str(err)
                ) from err
        if query.split_results:
            ms = es_dsl.MultiSearch(using=self.es)

            for resource in query.resources:
                s = es_dsl.Search(index=resource)

                if es_query is not None:
                    s = s.query(es_query)
                s = s[query.from_ : query.from_ + query.size]
                if query.sort:
                    s = s.sort(*self.translate_sort_fields([resource], query.sort))
                elif query.sort_dict and resource in query.sort_dict:
                    s = s.sort(
                        *self.translate_sort_fields(
                            [resource], query.sort_dict[resource]
                        )
                    )
                ms = ms.add(s)

            responses = ms.execute()
            result: dict[str, Any] = {"total": 0, "hits": {}}
            for i, response in enumerate(responses):
                result["hits"][query.resources[i]] = self._format_result(
                    query.resources, response
                ).get("hits", [])
                result["total"] += response.hits.total
                if query.lexicon_stats:
                    if "distribution" not in result:
                        result["distribution"] = {}
                    result["distribution"][query.resources[i]] = response.hits.total
            return result
        else:
            s = es_dsl.Search(using=self.es, index=query.resources, doc_type="entry")
            if es_query is not None:
                s = s.query(es_query)

            s = s[query.from_ : query.from_ + query.size]

            if query.lexicon_stats:
                s.aggs.bucket(
                    "distribution", "terms", field="_index", size=len(query.resources)
                )
            if query.sort:
                s = s.sort(*self.translate_sort_fields(query.resources, query.sort))
            elif query.sort_dict:
                sort_fields = []
                for resource, sort in query.sort_dict.items():
                    sort_fields.extend(self.translate_sort_fields([resource], sort))
                s = s.sort(*sort_fields)
            logger.debug("s = %s", extra={"es_query s": s.to_dict()})
            response = s.execute()

            # TODO format response in a better way, because the whole response takes up too much space in the logs
            # logger.debug('response = {}'.format(response.to_dict()))

            logger.debug("calling _format_result")
            result = self._format_result(query.resources, response)
            if query.lexicon_stats:
                result["distribution"] = {}
                for bucket in response.aggregations.distribution.buckets:
                    key = bucket["key"]
                    value = bucket["doc_count"]
                    result["distribution"][key.rsplit("_", 1)[0]] = value

            # logger.debug("return result = %s", result)
            return result

    def translate_sort_fields(
        self, resources: List[str], sort_values: List[str]
    ) -> List[Union[str, Dict[str, Dict[str, str]]]]:
        """Translate sort field to ES sort fields.

        Arguments:
            sort_values {List[str]} -- values to sort by

        Returns:
            List[str] -- values that ES can sort by.
        """
        translated_sort_fields: List[Union[str, Dict[str, Dict[str, str]]]] = []
        for sort_value in sort_values:
            sort_order = None
            if "|" in sort_value:
                sort_value, sort_order = sort_value.split("|", 1)
            for resource_id in resources:
                if sort_order:
                    translated_sort_fields.extend(
                        (
                            {field: {"order": sort_order}}
                            for field in self.translate_sort_field(
                                resource_id, sort_value
                            )
                        )
                    )
                translated_sort_fields.extend(
                    self.translate_sort_field(resource_id, sort_value)
                )

        return translated_sort_fields

    def translate_sort_field(self, resource_id: str, sort_value: str) -> List[str]:
        logger.debug(
            "sortable fields for resource",
            extra={
                "resource_id": resource_id,
                "sortable_fields": self.sortable_fields[resource_id],
            },
        )
        if sort_value in self.sortable_fields[resource_id]:
            return self.sortable_fields[resource_id][sort_value]
        else:
            raise UnsupportedField(
                f"You can't sort by field '{sort_value}' for resource '{resource_id}'"
            )

    def search_ids(self, resource_id: str, entry_ids: str):
        logger.info(
            "Called EsSearch.search_ids with:",
            extra={"resource_id": resource_id, "entry_ids": entry_ids},
        )
        entries = entry_ids.split(",")
        query = es_dsl.Q("terms", _id=entries)
        logger.debug("query", extra={"query": query})
        s = es_dsl.Search(using=self.es, index=resource_id).query(query)
        logger.debug("s", extra={"es_query s": s.to_dict()})
        response = s.execute()

        return self._format_result([resource_id], response)

    def statistics(self, resource_id: str, field: str) -> Iterable:
        s = es_dsl.Search(using=self.es, index=resource_id)
        s = s[0:0]

        if field in self.analyzed_fields[resource_id]:
            field += ".raw"

        logger.debug("Statistics: analyzed fields are:")
        logger.debug(json.dumps(self.analyzed_fields, indent=4))
        logger.debug(
            "Doing aggregations on resource_id: {resource_id}, on field {field}".format(
                resource_id=resource_id, field=field
            )
        )
        s.aggs.bucket("field_values", "terms", field=field, size=2147483647)
        response = s.execute()
        logger.debug("Elasticsearch response", extra={"response": response})
        return [
            {"value": bucket["key"], "count": bucket["doc_count"]}
            for bucket in response.aggregations.field_values.buckets
        ]

    def on_publish_resource(self, alias_name: str, index_name: str):
        mapping = self._get_index_mappings(index=index_name)
        if (
            "mappings" in mapping[index_name]
            and "entry" in mapping[index_name]["mappings"]
            and "properties" in mapping[index_name]["mappings"]["entry"]
        ):
            self.analyzed_fields[
                alias_name
            ] = Es6SearchService.get_analyzed_fields_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )
            self.sortable_fields[
                alias_name
            ] = Es6SearchService.create_sortable_map_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )

    @staticmethod
    def create_sortable_map_from_mapping(properties: dict) -> dict[str, list[str]]:
        sortable_map: dict[str, list[str]] = {}

        def parse_prop_value(sort_map, base_name, prop_name, prop_value: Dict):
            if "properties" in prop_value:
                for ext_name, ext_value in prop_value["properties"].items():
                    ext_base_name = f"{base_name}.{ext_name}"
                    parse_prop_value(sort_map, ext_base_name, ext_base_name, ext_value)
                return
            if prop_value["type"] in [
                "boolean",
                "date",
                "double",
                "keyword",
                "long",
                "ip",
            ]:
                sort_map[base_name] = [prop_name]
                sort_map[prop_name] = [prop_name]
                return
            if prop_value["type"] == "text":
                if "fields" in prop_value:
                    for ext_name, ext_value in prop_value["fields"].items():
                        parse_prop_value(
                            sort_map, base_name, f"{base_name}.{ext_name}", ext_value
                        )
                return

        for prop_name, prop_value in properties.items():
            parse_prop_value(sortable_map, prop_name, prop_name, prop_value)
            # if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
            #     sortable_map[prop_name] = prop_name
            # if prop_value["type"] == "text":
            #     if "fields" in prop_value:

        return sortable_map


# def parse_sortable_fields(properties: Dict[str, Any]) -> Dict[str, List[str]]:
#     for prop_name, prop_value in properties.items():
#         if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
#             return [prop_name]



import typing

from karp.lex.application.repositories import ResourceUnitOfWork
from karp.search.application.queries import SearchQuery
from karp.search.application.repositories import SearchServiceUnitOfWork
from karp.search.domain import search_service, errors


class GenericSearchQuery(SearchQuery):
    def __init__(
        self,
        resource_uow: ResourceUnitOfWork,
        search_service_uow: SearchServiceUnitOfWork,
    ):
        self.resource_uow = resource_uow
        self.search_service_uow = search_service_uow

    def check_resource_published(self, resource_id: str):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(resource_id)
            if not resource.is_published:
                raise errors.ResourceNotPublished(resource_id)

    def check_all_resources_published(
        self,
        resource_ids: typing.Iterable[str]
    ):
        with self.resource_uow:
            for resource_id in resource_ids:
                resource = self.resource_uow.repo.by_resource_id(resource_id)
                if not resource.is_published:
                    raise errors.ResourceNotPublished(resource_id)

    def search_ids(self, resource_id: str, entry_ids: str):
        self.check_resource_published(resource_id)
        with self.search_service_uow:
            return self.search_service_uow.repo.search_ids(resource_id, entry_ids)

    def query(self, req: search_service.QueryRequest):
        print(f"entry_query.query called with req={req}")
        self.check_all_resources_published(req.resource_ids)

        with self.search_service_uow:
            return self.search_service_uow.repo.query(req)

    def query_split(self, req: search_service.QueryRequest):
        self.check_all_resources_published(req.resource_ids)
        with self.search_service_uow as uw:
            return uw.repo.query_split(req)

    def statistics(self, resource_id: str, field: str):
        self.check_resource_published(resource_id)
        with self.search_service_uow as uw:
            return uw.repo.statistics(resource_id, field)



import abc
import logging
from typing import Dict, List, Type, Union, TypeVar, Generic, Iterable

from .commands import Command
from .events import Event


logger = logging.getLogger(__name__)

T = TypeVar("T")


class Handler(Generic[T]):
    @abc.abstractmethod
    def execute(self, t: T) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def collect_new_events(self) -> Iterable[Event]:
        raise NotImplementedError


CommandType = TypeVar('CommandType', bound=Command)
EventType = TypeVar('EventType', bound=Event)


class CommandHandler(Generic[CommandType], Handler[CommandType]):
    pass


Message = Union[Command, Event]


class MessageBus:
    def __init__(
        self,
        command_handlers: Dict[Type[Command], CommandHandler[Command]],
        event_handlers: Dict[Type[Event], List[Handler[Event]]],
        raise_on_all_errors: bool = False,
    ) -> None:
        self._command_handlers = command_handlers
        self._event_handlers = event_handlers
        self.raise_on_all_errors = raise_on_all_errors
        self.queue: List[Message] = []

    def handle(self, msg: Message):
        self.queue = [msg]
        while self.queue:
            msg = self.queue.pop(0)
            if isinstance(msg, Event):
                self._handle_event(msg)
            elif isinstance(msg, Command):
                self._handle_command(msg)
            else:
                raise Exception(f"{msg} was not an Event or Command")

    def _handle_event(self, event: Event):
        for evt_handler in self._event_handlers[type(event)]:
            logger.debug("handling event %s with handler %s",
                         event, evt_handler
                         )
            try:
                evt_handler.execute(event)
            except Exception:
                logger.exception("Exception handling event %s", event)
                if self.raise_on_all_errors:
                    raise
            else:
                self.queue.extend(evt_handler.collect_new_events())

    def _handle_command(self, command: Command):
        logger.debug("handling command %s", command)
        cmd_handler = self._command_handlers[type(command)]
        try:
            cmd_handler.execute(command)
        except Exception:
            logger.exception("Exception handling command %s", command)
            raise
        self.queue.extend(cmd_handler.collect_new_events())



from typing import Iterable, Optional

import sqlalchemy as sa
from sqlalchemy import sql
from karp.foundation.value_objects.unique_id import UniqueId

from karp.lex.application.queries import (
    GetPublishedResources,
    ResourceDto,
    GetResources,
)
from karp.lex.application.queries.resources import ReadOnlyResourceRepository
from karp.lex.domain.entities import resource
from karp.lex_infrastructure.sql.sql_models import ResourceModel
from karp.lex_infrastructure.queries.base import SqlQuery


class SqlGetPublishedResources(GetPublishedResources, SqlQuery):
    def query(self) -> Iterable[ResourceDto]:
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.is_published == True,
            ),
        )
        return (_row_to_dto(row) for row in self._conn.execute(stmt))


class SqlGetResources(GetResources, SqlQuery):
    def query(self) -> Iterable[ResourceDto]:
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
            ),
        )
        return (_row_to_dto(row) for row in self._conn.execute(stmt))


class SqlReadOnlyResourceRepository(ReadOnlyResourceRepository, SqlQuery):
    def get_by_id(
        self, entity_id: UniqueId, version: Optional[int] = None
    ) -> Optional[ResourceDto]:
        filters: dict[str, UniqueId | str | int] = {"entity_id": entity_id}
        if version:
            filters["version"] = version
        stmt = (
            sql.select(ResourceModel)
            .filter_by(**filters)
            .order_by(ResourceModel.last_modified.desc())
        )
        print(f"stmt={str(stmt)}")
        row = self._conn.execute(stmt).first()

        return _row_to_dto(row) if row else None

    def _get_by_resource_id(self, resource_id: str) -> Optional[ResourceDto]:
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.resource_id == resource_id,
            ),
        )
        stmt = stmt.order_by(ResourceModel.last_modified.desc())
        row = self._conn.execute(stmt).first()

        return _row_to_dto(row) if row else None

    def get_published_resources(self) -> Iterable[ResourceDto]:
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.is_published == True,
            ),
        )
        return (_row_to_dto(row) for row in self._conn.execute(stmt))

    def get_all_resources(self) -> Iterable[ResourceDto]:
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
            ),
        )
        return (_row_to_dto(row) for row in self._conn.execute(stmt))


def _row_to_dto(row_proxy) -> ResourceDto:
    return ResourceDto(
        entity_id=row_proxy.entity_id,
        resource_id=row_proxy.resource_id,
        version=row_proxy.version,
        config=row_proxy.config,
        is_published=row_proxy.is_published,
        entry_repository_id=row_proxy.entry_repo_id,
        last_modified=row_proxy.last_modified,
        last_modified_by=row_proxy.last_modified_by,
        message=row_proxy.message,
        name=row_proxy.name,
        discarded=row_proxy.discarded,
    )



from fastapi import APIRouter

from karp.webapp.routes import (
    entries_api,
    health_api,
    history_api,
    query_api,
    resources_api,
    stats_api,
)


router = APIRouter()


router.include_router(
    entries_api.router,
    prefix='/entries',
    tags=['Entries'],
)
router.include_router(health_api.router, prefix='/healthz')
router.include_router(
    history_api.router,
    prefix='/history',
    tags=['Entries', 'History'],
)
router.include_router(
    query_api.router,
    prefix='/query',
    tags=['Querying'],
)
router.include_router(
    resources_api.router,
    prefix='/resources',
    tags=['Resources'],
)
router.include_router(
    stats_api.router,
    prefix='/stats',
    tags=['Statistics'],
)



import typing
from enum import Enum
from typing import Dict, List, Optional

import pydantic

from karp.search.domain import errors, query_dsl  # , resourcemgr
# from karp.domain import errors
from karp.util import convert as util_convert
from karp.utility.container import arg_get


class Format(str, Enum):
    json = "json"
    csv = "csv"
    xml = "xml"
    lmf = "lmf?"
    tsb = "tsb"


class Query(pydantic.BaseModel):
    fields: typing.List[str]
    resources: typing.List[str]
    sort: typing.List[str]
    from_: int = pydantic.Field(0, alias="from")
    size: int = 25
    split_results: bool = False
    lexicon_stats: bool = True
    include_fields: typing.Optional[typing.List[str]] = None
    exclude_fields: typing.Optional[typing.List[str]] = None
    format_: typing.Optional[Format] = pydantic.Field(None, alias="format")
    format_query: typing.Optional[Format] = None
    q: typing.Optional[str] = None
    ast: typing.Optional[query_dsl.Ast] = None
    sort_dict: typing.Optional[typing.Dict[str, typing.List[str]]] = pydantic.Field(
        default_factory=dict
    )

    @pydantic.validator(
        "resources", "include_fields", "exclude_fields", "sort", pre=True
    )
    @classmethod
    def split_str(cls, v):
        if isinstance(v, str):
            return v.split(",")
        return v

    @pydantic.validator("fields", "sort", pre=True, always=True)
    @classmethod
    def set_ts_now(cls, v):
        return v or []

    class Config:
        arbitrary_types_allowed = True

    def parse_arguments(self, args, resource_str: str):
        if resource_str is None:
            raise errors.IncompleteQuery("No resources are defined.")
        self.resources = resource_str.split(",")
        self.from_ = arg_get(args, "from", int, 0)
        self.size = arg_get(args, "size", int, 25)
        self.lexicon_stats = arg_get(
            args, "lexicon_stats", util_convert.str2bool, True)
        self.include_fields = arg_get(
            args, "include_fields", util_convert.str2list(",")
        )
        self.exclude_fields = arg_get(
            args, "exclude_fields", util_convert.str2list(",")
        )
        self.fields = []
        self.format = arg_get(args, "format")
        self.format_query = arg_get(args, "format_query")
        self.q = arg_get(args, "q") or ""
        self.sort: List[str] = arg_get(
            args, "sort", util_convert.str2list(",")) or []
        # self.sort_dict: Dict[str, List[str]] = {}
        #         if not self.sort:
        #             if len(self.resources) == 1:
        #                 self.sort = resourcemgr.get_resource(self.resources[0]).default_sort()
        #             else:
        #                 for resource_id in self.resources:
        #                     self.sort_dict[resource_id] = resourcemgr.get_resource(
        #                         resource_id
        #                     ).default_sort()
        self.ast = query_dsl.parse(self.q)
        self._update_ast()

    def _update_ast(self):
        if self.ast.is_empty():
            return

        field_translations = {}
        for resource in self.resources:
            # ft = resourcemgr.get_field_translations(resource)
            ft = {}
            if ft:
                for field, lst in ft.items():
                    if field in field_translations:
                        field_translations[field].update(lst)
                    else:
                        field_translations[field] = set(lst)

        def translate_node(node: query_dsl.Node):
            print("node = {node!r}".format(node=node))
            if query_dsl.is_a(
                node, [query_dsl.op.FREERGXP,
                       query_dsl.op.FREETEXT, query_dsl.op.ARGS]
            ):
                return

            if query_dsl.is_a(node, query_dsl.op.LOGICAL):
                for child in node.children:
                    translate_node(child)
            elif query_dsl.is_a(node, query_dsl.op.OPS):
                print("|OPS| node.children = {}".format(node.children))
                field = node.children[0]
                if query_dsl.is_a(field, query_dsl.op.STRING):
                    if field.value in field_translations:
                        fields = query_dsl.Node(query_dsl.op.ARG_OR, None)
                        fields.add_child(field)
                        for _ft in field_translations[field.value]:
                            fields.add_child(
                                query_dsl.Node(query_dsl.op.STRING, 0, _ft)
                            )
                        node.children[0] = fields
                else:
                    translate_node(field)
            # elif query_dsl.is_a(node, query_dsl.op.ARG_LOGICAL):
            elif query_dsl.is_a(node, query_dsl.op.ARG_OR):
                print(
                    "|ARG_OR| node.children = {node.children}".format(node=node))

                for child in node.children:
                    if child.value in field_translations:
                        for _ft in field_translations[child.value]:
                            node.add_child(query_dsl.Node(
                                query_dsl.op.STRING, 0, _ft))
            elif query_dsl.is_a(node, query_dsl.op.ARG_LOGICAL):  # ARG_OR handled above
                print(
                    "|ARG_AND| node.children = {node.children}".format(node=node))
                changes = []
                for i, child in enumerate(node.children):
                    if child.value in field_translations:
                        fields = query_dsl.Node(query_dsl.op.ARG_OR, None)
                        fields.add_child(child)
                        for _ft in field_translations[child.value]:
                            fields.add_child(
                                query_dsl.Node(query_dsl.op.STRING, 0, _ft)
                            )
                        changes.append((i, fields))
                print("|ARG_AND| changes = {changes}".format(changes=changes))
                for i, fields in changes:
                    node.children[i] = fields
                print(
                    "|ARG_AND| node.children = {node.children}".format(node=node))

        translate_node(self.ast.root)
        # TODO rewrite

    # def __repr__(self) -> str:
    #     return """
    #         {}(
    #             q={}
    #             resources={}
    #             include_fields={}
    #             exclude_fields={}
    #             fields={}
    #             sort={}
    #             from={}, size={},
    #             split_results={}, lexicon_stats={},
    #             format={}
    #             format_query={}
    #         )""".format(
    #         self._self_name(),
    #         self.q,
    #         self.resources,
    #         self.include_fields,
    #         self.exclude_fields,
    #         self.fields,
    #         self.sort,
    #         self.from_,
    #         self.size,
    #         self.split_results,
    #         self.lexicon_stats,
    #         self.format,
    #         self.format_query,
    #     )

    def _self_name(self) -> str:
        return "Query"



import typing

import pydantic


class AccessToken(pydantic.BaseModel):
    access_token: str
    token_type: str

    def as_header(self) -> typing.Dict[str, str]:
        return {
            'Authorization': f'{self.token_type} {self.access_token}',
        }



import injector
import pytest

from karp.foundation.commands import CommandBus
from karp.foundation.events import EventBus
from karp.lex import Lex
from karp.lex_infrastructure import GenericLexInfrastructure
from karp.main.modules import CommandBusMod, EventBusMod
from karp.search import Search
from karp.search_infrastructure import SearchInfrastructure, GenericSearchInfrastructure
from karp.tests.unit.lex.adapters import InMemoryLexInfrastructure

from . import adapters


@pytest.fixture()
def search_unit_ctx() -> adapters.SearchUnitTestContext:
    container = injector.Injector([
        CommandBusMod(),
        EventBusMod(),
        Search(),
        SearchInfrastructure(),
        Lex(),
        GenericLexInfrastructure(),
        InMemoryLexInfrastructure(),
        GenericSearchInfrastructure(),
        adapters.InMemorySearchInfrastructure(),
    ], auto_bind=False)
    return adapters.SearchUnitTestContext(
        container=container,
        command_bus=container.get(CommandBus),  # type: ignore
        event_bus=container.get(EventBus),  # type: ignore
    )



import abc
import logging
from typing import List, Optional

from karp.foundation import value_objects
from karp.auth.domain import errors
from karp.auth.domain.entities.user import User


logger = logging.getLogger("karp")

class AuthServiceConfig:
    pass


class AuthService(abc.ABC):

    @abc.abstractmethod
    def authenticate(self, scheme: str, credentials: str) -> User:
        return User("dummy", {}, {})

    @abc.abstractmethod
    def authorize(
        self,
        level: value_objects.PermissionLevel,
        user: User,
        resource_ids: List[str],
    ) -> bool:
        return True



from pathlib import Path

from alembic.config import Config as AlembicConfig
import alembic
import pydantic

from karp.foundation import commands


class RunMigrationsUp(pydantic.BaseModel):
    to_revision: str = 'head'


class RunMigrationsDown(pydantic.BaseModel):
    to_revision: str = 'base'


class RunningMigrationsBase:
    def __init__(self):
        self.alembic_cfg = AlembicConfig(
            Path(__file__).parent / 'alembic.ini'
        )
        self.alembic_cfg.set_main_option(
            'script_location',
            str(Path(__file__).parent)
        )


class RunningMigrationsUp(
    RunningMigrationsBase,
    commands.CommandHandler[RunMigrationsUp],
):
    def execute(self, command: RunMigrationsUp):
        alembic.command.upgrade(
            self.alembic_cfg,
            command.to_revision,
        )


class RunningMigrationsDown(
    RunningMigrationsBase,
    commands.CommandHandler[RunMigrationsDown]
):
    def execute(self, command: RunMigrationsDown):
        alembic.command.downgrade(
            self.alembic_cfg,
            command.to_revision,
        )



from typing import Any

import pytest
from karp.foundation.value_objects.unique_id import make_unique_id

from karp.lex import EntryDto
from karp.search import EntryTransformer
from karp.search_infrastructure import GenericEntryTransformer

from karp.tests.unit.lex import factories as lex_factories


class TestGenericEntryTransformer:
    @pytest.mark.parametrize(
        "field_name, field_config, field_value",
        [
            ("single", {"type": "boolean"}, True),
        ],
    )
    def test_transform_to_index_entry(
        self,
        field_name: str,
        field_config: dict,
        field_value: Any,
        search_unit_ctx,
    ):
        resource_id = "transform_res"
        create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
        search_unit_ctx.command_bus.dispatch(create_entry_repo)
        create_resource = lex_factories.CreateResourceFactory(
            entry_repo_id=create_entry_repo.entity_id,
            resource_id=resource_id,
            config={
                "fields": {
                    "id": {"type": "string"},
                    field_name: field_config,
                },
                "id": "id",
            },
        )
        search_unit_ctx.command_bus.dispatch(create_resource)
        transformer = search_unit_ctx.container.get(EntryTransformer)  # type: ignore [misc]
        entry_id = "entry..1"
        src_entry = EntryDto(
            entity_id=make_unique_id(),
            resource=resource_id,
            version=1,
            entry={"id": entry_id, field_name: field_value},
            last_modified=1234567,
            last_modified_by="alice@example.com",
        )
        index_entry = transformer.transform(resource_id, src_entry)
        assert index_entry.id == str(src_entry.entity_id)
        assert index_entry.entry["_entry_version"] == 1
        assert index_entry.entry["id"] == entry_id
        assert index_entry.entry[field_name] == field_value

    @pytest.mark.parametrize(
        "field_name, field_config, field_value",
        [
            ("single", {"type": "boolean"}, True),
            ("single", {"type": "string"}, "plain string"),
            ("single", {"type": "number"}, 3.14),
            ("single", {"type": "integer"}, 3),
            ("single", {"type": "long_string"}, "very long string"),
            (
                "single",
                {
                    "type": "object",
                    "fields": {
                        "sub": {"type": "string"},
                    },
                },
                {"sub": "test"},
            ),
        ],
    )
    def test_transform_to_index_entry_collection(
        self,
        field_name: str,
        field_config: dict,
        field_value: Any,
        search_unit_ctx,
    ):
        resource_id = "transform_res"
        create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
        search_unit_ctx.command_bus.dispatch(create_entry_repo)
        field_config["collection"] = True
        create_resource = lex_factories.CreateResourceFactory(
            entry_repo_id=create_entry_repo.entity_id,
            resource_id=resource_id,
            config={
                "fields": {
                    "id": {"type": "string"},
                    field_name: field_config,
                },
                "id": "id",
            },
        )
        search_unit_ctx.command_bus.dispatch(create_resource)
        transformer = search_unit_ctx.container.get(EntryTransformer)  # type: ignore [misc]
        entry_id = "entry..1"
        src_entry = EntryDto(
            entity_id=make_unique_id(),
            resource=resource_id,
            version=1,
            entry={
                "id": entry_id,
                field_name: [field_value],
            },
            last_modified=1234567,
            last_modified_by="alice@example.com",
        )
        index_entry = transformer.transform(resource_id, src_entry)
        assert index_entry.id == str(src_entry.entity_id)
        assert index_entry.entry["_entry_version"] == 1
        assert index_entry.entry["id"] == entry_id
        assert index_entry.entry[field_name][0] == field_value

    @pytest.mark.parametrize(
        "field_name, field_config, field_value",
        [
            ("single", {"type": "boolean"}, True),
            ("single", {"type": "string"}, "plain string"),
            ("single", {"type": "number"}, 3.14),
            ("single", {"type": "integer"}, 3),
            ("single", {"type": "long_string"}, "very long string"),
            (
                "single",
                {"type": "object", "fields": {"sub": {"type": "string"}}},
                {"sub": "test"},
            ),
        ],
    )
    def test_transform_to_index_entry_object(
        self,
        field_name: str,
        field_config: dict,
        field_value: Any,
        search_unit_ctx,
    ):
        resource_id = "transform_res"
        create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
        search_unit_ctx.command_bus.dispatch(create_entry_repo)
        create_resource = lex_factories.CreateResourceFactory(
            entry_repo_id=create_entry_repo.entity_id,
            resource_id=resource_id,
            config={
                "fields": {
                    "id": {"type": "string"},
                    "obj": {
                        "type": "object",
                        "fields": {
                            field_name: field_config,
                        },
                    },
                },
                "id": "id",
            },
        )
        search_unit_ctx.command_bus.dispatch(create_resource)
        transformer = search_unit_ctx.container.get(EntryTransformer)  # type: ignore [misc]
        entry_id = "entry..1"
        src_entry = EntryDto(
            entity_id=make_unique_id(),
            resource=resource_id,
            version=1,
            entry={
                "id": entry_id,
                "obj": {
                    field_name: field_value,
                },
            },
            last_modified=1234567,
            last_modified_by="alice@example.com",
        )
        index_entry = transformer.transform(resource_id, src_entry)
        assert index_entry.id == str(src_entry.entity_id)
        assert index_entry.entry["_entry_version"] == 1
        assert index_entry.entry["id"] == entry_id
        assert index_entry.entry["obj"][field_name] == field_value



import pytest

from fastapi import status


class TestQueryApiInput:
    @pytest.mark.parametrize("query_string", [
        "sort=name",
        "sort=name|asc",
        "sort=name|desc",
        "from=10",
        "size=10",
        "q=",
        "q=&sort=name&from=0&size=25",
    ])
    @pytest.mark.parametrize("path", ["query", "query/split"])
    def test_valid_query_strings(
        self, 
        fa_client, 
        path: str,
        query_string: str
    ):
        response = fa_client.get(f"/{path}/places?{query_string}")

        print(f"{response.json()=}")
        assert response.status_code != status.HTTP_422_UNPROCESSABLE_ENTITY




from starlette.config import Config as StarletteConfig
from starlette.datastructures import CommaSeparatedStrings

config = StarletteConfig(".env")


ELASTICSEARCH_HOST = config(
    "ELASTICSEARCH_HOST", cast=CommaSeparatedStrings, default=None
)



import typing

from fastapi import (APIRouter, Depends, HTTPException, Response, Security,
                     status)

from karp import auth
from karp.foundation.value_objects import PermissionLevel
from karp.search.application.queries import SearchService, StatisticsDto
from karp.webapp import schemas

from karp.webapp import dependencies as deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req


router = APIRouter()


@router.get(
    "/{resource_id}/{field}",
    response_model=typing.List[StatisticsDto],
)
def get_field_values(
    resource_id: str,
    field: str,
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    if not auth_service.authorize(PermissionLevel.read, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    print(f"calling statistics ... from {search_service=}")
    return search_service.statistics(resource_id, field)


def init_app(app):
    app.include_router(router)



from typing import Any, Callable, Dict, List, Optional, TypeVar

T = TypeVar("T", bool, int, str, List[str])


def arg_get(
    args: Dict,
    arg_name: str,
    convert: Optional[Callable[[str], T]] = None,
    default: Optional[T] = None,
) -> Optional[T]:
    arg = args.get(arg_name, None)
    if arg is None:
        return default
    if convert is None:
        return arg
    return convert(arg)


def create_field_getter(
    field: str,
    convert: Optional[Callable[[Any], Any]] = None
) -> Callable[[Dict], Any]:
    def getter(d: Dict):
        result = d[field]
        return convert(result) if convert else result

    return getter



import pytest
from fastapi import status
from karp import auth


def test_stats_wo_auth(fa_data_client):
    response = fa_data_client.get(
        "/stats/places/area",
    )
    assert response.status_code == status.HTTP_200_OK

    entries = response.json()
    print(f'{entries=}')
    assert len(entries) == 4


def test_stats_w_auth(fa_data_client, read_token: auth.AccessToken):
    response = fa_data_client.get(
        "/stats/places/area",
        headers=read_token.as_header(),
    )
    assert response.status_code == status.HTTP_200_OK

    entries = response.json()
    print(f'{entries=}')
    assert len(entries) == 4



import abc
from typing import Dict, Optional, Protocol

import injector

from karp.foundation import repository, unit_of_work, events
from karp.foundation.value_objects import UniqueId
from karp.lex.domain import errors
from .entries import EntryUnitOfWork


class EntryRepositoryUnitOfWorkFactory:
    @abc.abstractmethod
    def create(
        self,
        repository_type: str,
        entity_id: UniqueId,
        name: str,
        config: dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> EntryUnitOfWork:
        pass


class EntryUnitOfWorkCreator(Protocol):
    def __call__(
        self,
        entity_id: UniqueId,
        name: str,
        config: dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> EntryUnitOfWork:
        ...


class InjectorEntryUnitOfWorkRepoFactory(
    EntryRepositoryUnitOfWorkFactory
):
    def __init__(self, container: injector.Injector) -> None:
        self._container = container

    def create(
        self,
        repository_type: str,
        entity_id: UniqueId,
        name: str,
        config: dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> EntryUnitOfWork:
        uow_factory_cls = self._container.get(
            Dict[str, EntryUnitOfWorkCreator]
        )[repository_type]

        uow_factory: EntryUnitOfWorkCreator = self._container.create_object(
            uow_factory_cls)
        return uow_factory(
            entity_id=entity_id,
            name=name,
            config=config,
            connection_str=connection_str,
            user=user,
            message=message,
            timestamp=timestamp,
        )


class EntryUowRepository(repository.Repository[EntryUnitOfWork]):
    EntityNotFound = errors.EntryRepoNotFound
    pass


class EntryUowRepositoryUnitOfWork(unit_of_work.UnitOfWork[EntryUowRepository]):
    def __init__(
        self,
        *,
        event_bus: events.EventBus,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
    ):
        unit_of_work.UnitOfWork.__init__(self, event_bus)
        self._entry_uow_factory = entry_uow_factory

    @property
    def factory(self) -> EntryRepositoryUnitOfWorkFactory:
        return self._entry_uow_factory



# from elasticsearch import Elasticsearch  # pyre-ignore

# from karp.application import ctx
# from karp.infrastructure.elasticsearch6.es6_search_service import Es6SearchService


# def init_es(host):
#     print("Setting up ES with host={}".format(host))
#     es = Elasticsearch(
#         hosts=host,
#         sniff_on_start=True,
#         sniff_on_connection_fail=True,
#         sniffer_timeout=60,
#         sniff_timeout=10,
#     )
#     search_service = Es6SearchService(es)
#     ctx.search_service = search_service



import abc
import typing

from karp.search.application.repositories import IndexEntry


class PreProcessor(abc.ABC):
    @abc.abstractmethod
    def process(self, resource_id: str) -> typing.Iterable[IndexEntry]:
        raise NotImplementedError()




"""use ulid as str

Revision ID: c9a01aaad322
Revises: d2f8d1c19b65
Create Date: 2022-12-01 12:36:55.684153

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy_utils import UUIDType
from karp.db_infrastructure.types import ULIDType


# revision identifiers, used by Alembic.
revision = "c9a01aaad322"
down_revision = "d2f8d1c19b65"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("resources", "entity_id", nullable=False, type_=ULIDType)
    op.alter_column("resources", "entry_repo_id", nullable=False, type_=ULIDType)
    op.alter_column("entry_repos", "entity_id", nullable=False, type_=ULIDType)


def downgrade():
    op.alter_column("resources", "entity_id", nullable=False, type_=UUIDType)
    op.alter_column("resources", "entry_repo_id", nullable=False, type_=UUIDType)
    op.alter_column("entry_repos", "entity_id", nullable=False, type_=UUIDType)



from pathlib import Path
import injector

from karp.auth import (
    AuthService,
    AuthServiceConfig,
    GetResourcePermissions,
    IsResourceProtected,
)
from karp.auth_infrastructure.queries import LexGetResourcePermissions, LexIsResourceProtected
from karp.auth_infrastructure.services import DummyAuthService, JWTAuthService, JWTAuthServiceConfig
from karp.lex.application.queries import GetPublishedResources, ReadOnlyResourceRepository


class AuthInfrastructure(injector.Module):
    @injector.provider
    def resource_permissions(
        self,
        get_published_resources: GetPublishedResources
    ) -> GetResourcePermissions:
        return LexGetResourcePermissions(get_published_resources)

    @injector.provider
    def is_resource_protected(
        self,
        resource_repo: ReadOnlyResourceRepository
    ) -> IsResourceProtected:
        return LexIsResourceProtected(resource_repo)


class TestAuthInfrastructure(injector.Module):
    @injector.provider
    def dummy_auth_service(self) -> AuthService:
        return DummyAuthService()


class JwtAuthInfrastructure(injector.Module):
    def __init__(self, pubkey_path: Path) -> None:
        super().__init__()
        self.pubkey_path = pubkey_path

    @injector.provider
    @injector.singleton
    def jwt_auth_service_config(self) -> AuthServiceConfig:
        return JWTAuthServiceConfig(self.pubkey_path)

    @injector.provider
    def jwt_auth_service(
        self,
        is_resource_protected: IsResourceProtected
    ) -> AuthService:
        return JWTAuthService(
            pubkey_path=self.pubkey_path,
            is_resource_protected=is_resource_protected
        )



"""Commands top-level module."""

from .entry_commands import (
    AddEntries,
    AddEntriesInChunks,
    AddEntry,
    DeleteEntry,
    ImportEntries,
    ImportEntriesInChunks,
    UpdateEntry,
)
from .entry_repo_commands import CreateEntryRepository, DeleteEntryRepository
from .resource_commands import (
    CreateResource,
    DeleteResource,
    PublishResource,
    SetEntryRepoId,
    UpdateResource,
)

__all__ = [
    "AddEntriesInChunks",
    "ImportEntries",
    "ImportEntriesInChunks",
]



import logging
from typing import Callable
from fastapi import FastAPI

from karp.db_infrastructure import tasks as db_tasks
from karp.main import config


logger = logging.getLogger(__name__)


def create_start_app_handler(app: FastAPI) -> Callable:
    async def start_app() -> None:
        logger.debug('start_app')
        app.state.db = db_tasks.connect_to_db(config.DATABASE_URL)
    return start_app


def create_stop_app_handler(app: FastAPI) -> Callable:
    async def stop_app() -> None:
        logger.debug('stop_app')
        db_tasks.close_db_connection(app.state.db)
        app.state.app_context = None
    return stop_app




import abc
import logging
from typing import Any, Generic, TypeVar

import injector


CommandType = TypeVar("CommandType")


logger = logging.getLogger(__name__)


class CommandHandler(Generic[CommandType]):
    def execute(self, command: CommandType) -> Any:
        raise NotImplementedError()


class Command:
    pass


class CommandBus(abc.ABC):
    @abc.abstractmethod
    def dispatch(self, command: Command) -> Any:
        raise NotImplementedError


class InjectorCommandBus(CommandBus):
    def __init__(self, container: injector.Injector) -> None:
        self._container = container

    def dispatch(self, command: Command) -> Any:
        logger.info("Handling command: %s", command)
        cmd_cls = type(command)
        cmd_handler = self._container.get(CommandHandler[cmd_cls])  # type: ignore

        logger.info("Handling command %s with handler %s", command, cmd_handler)
        return cmd_handler.execute(command)



from karp.foundation.value_objects import UniqueId
from karp.lex.application.queries import GetEntryRepositoryId
from karp.lex.application.repositories import ResourceUnitOfWork


class GenericGetEntryRepositoryId(GetEntryRepositoryId):
    def __init__(self, resource_uow: ResourceUnitOfWork) -> None:
        super().__init__()
        self._resource_uow = resource_uow

    def query(self, resource_id: str) -> UniqueId:
        with self._resource_uow as uw:
            return uw.repo.by_resource_id(resource_id).entry_repository_id



from itertools import zip_longest

import pytest  # pyre-ignore

from karp.search.domain.query_dsl import op, parser


def _test_nodes(r, facit):
    for x, f in zip_longest(r.gen_stream(), facit):
        assert x is not None, "x is too short"
        assert f is not None, "x is too long"

        assert x.type == f[0]
        assert x.value == f[1]
        assert isinstance(x.value, type(f[1]))


@pytest.mark.parametrize(
    "query,facit",
    [
        (
            "and||missing|pos||equals|wf||or|blomma|äpple",
            [
                (op.AND, None),
                (op.MISSING, None),
                (op.STRING, "pos"),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.ARG_OR, None),
                (op.STRING, "blomma"),
                (op.STRING, "äpple"),
            ],
        ),
        ("freetext|stort hus", [
         (op.FREETEXT, None), (op.STRING, "stort hus")]),
        (
            "freetext||not|stort hus",
            [(op.FREETEXT, None), (op.ARG_NOT, None), (op.STRING, "stort hus")],
        ),
        ("freetext|flicka", [(op.FREETEXT, None), (op.STRING, "flicka")]),
        (
            "freetext||and|3|flicka",
            [
                (op.FREETEXT, None),
                (op.ARG_AND, None),
                (op.INT, 3),
                (op.STRING, "flicka"),
            ],
        ),
        (
            "freetext||or|3|flicka",
            [
                (op.FREETEXT, None),
                (op.ARG_OR, None),
                (op.INT, 3),
                (op.STRING, "flicka"),
            ],
        ),
        ("freetext|6", [(op.FREETEXT, None), (op.INT, 6)]),
        (
            "startswith|lemgram|dalinm--",
            [(op.STARTSWITH, None), (op.STRING, "lemgram"), (op.STRING, "dalinm--")],
        ),
        (
            "startswith|lemgram||or|3|dalinm--",
            [
                (op.STARTSWITH, None),
                (op.STRING, "lemgram"),
                (op.ARG_OR, None),
                (op.STRING, "3"),
                (op.STRING, "dalinm--"),
            ],
        ),
        (
            "and||equals|wf|äta||missing|pos",
            [
                (op.AND, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "äta"),
                (op.MISSING, None),
                (op.STRING, "pos"),
            ],
        ),
        (
            "regexp|wf|.*o.*a",
            [(op.REGEXP, None), (op.STRING, "wf"), (op.STRING, ".*o.*a")],
        ),
        ("exists|sense", [(op.EXISTS, None), (op.STRING, "sense")]),
        ("not||exists|sense", [("NOT", None),
         (op.EXISTS, None), (op.STRING, "sense")]),
        (
            "and||equals|wf|sitta||not||equals|wf|satt",
            [
                (op.AND, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "sitta"),
                (op.NOT, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "satt"),
            ],
        ),
        ("freergxp|str.*ng", [(op.FREERGXP, None), (op.STRING, "str.*ng")]),
        (
            "freergxp||or|str.*ng1|str.*ng2",
            [
                (op.FREERGXP, None),
                (op.ARG_OR, None),
                (op.STRING, "str.*ng1"),
                (op.STRING, "str.*ng2"),
            ],
        ),
        (
            "regexp|field||or|str.*ng1|str.*ng2",
            [
                (op.REGEXP, None),
                (op.STRING, "field"),
                (op.ARG_OR, None),
                (op.STRING, "str.*ng1"),
                (op.STRING, "str.*ng2"),
            ],
        ),
        (
            "regexp||or|field1|field2||str.*ng",
            [
                (op.REGEXP, None),
                (op.ARG_OR, None),
                (op.STRING, "field1"),
                (op.STRING, "field2"),
                (op.STRING, "str.*ng"),
            ],
        ),
        (
            "exists||or|and|or",
            [
                (op.EXISTS, None),
                (op.ARG_OR, None),
                (op.STRING, "and"),
                (op.STRING, "or"),
            ],
        ),
        ("missing|field", [(op.MISSING, None), (op.STRING, "field")]),
        ("missing|3", [(op.MISSING, None), (op.STRING, "3")]),
        (
            "equals|field|string",
            [(op.EQUALS, None), (op.STRING, "field"), (op.STRING, "string")],
        ),
        (
            "equals|field|6.7",
            [(op.EQUALS, None), (op.STRING, "field"), (op.FLOAT, 6.7)],
        ),
        (
            "and||freergxp|str.*ng||regexp|field|str.*ng",
            [
                (op.AND, None),
                (op.FREERGXP, None),
                (op.STRING, "str.*ng"),
                (op.REGEXP, None),
                (op.STRING, "field"),
                (op.STRING, "str.*ng"),
            ],
        ),
        (
            "and||freergxp|str.*ng||regexp|field|str.*ng||exists|pos",
            [
                (op.AND, None),
                (op.FREERGXP, None),
                (op.STRING, "str.*ng"),
                (op.REGEXP, None),
                (op.STRING, "field"),
                (op.STRING, "str.*ng"),
                (op.EXISTS, None),
                (op.STRING, "pos"),
            ],
        ),
        (
            "or||freergxp|str.*ng||regexp|field|str.*ng",
            [
                (op.OR, None),
                (op.FREERGXP, None),
                (op.STRING, "str.*ng"),
                (op.REGEXP, None),
                (op.STRING, "field"),
                (op.STRING, "str.*ng"),
            ],
        ),
        (
            "and||regexp||or|wordform|baseform||s.tt.?||exists|pos",
            [
                (op.AND, None),
                (op.REGEXP, None),
                (op.ARG_OR, None),
                (op.STRING, "wordform"),
                (op.STRING, "baseform"),
                (op.STRING, "s.tt.?"),
                (op.EXISTS, None),
                (op.STRING, "pos"),
            ],
        ),
        (
            "equals|field||or|string1|string2|string3",
            [
                (op.EQUALS, None),
                (op.STRING, "field"),
                (op.ARG_OR, None),
                (op.STRING, "string1"),
                (op.STRING, "string2"),
                (op.STRING, "string3"),
            ],
        ),
        (
            "contains|field|string",
            [(op.CONTAINS, None), (op.STRING, "field"), (op.STRING, "string")],
        ),
        (
            "endswith|field|string",
            [(op.ENDSWITH, None), (op.STRING, "field"), (op.STRING, "string")],
        ),
        (
            "endswith|field|56",
            [(op.ENDSWITH, None), (op.STRING, "field"), (op.STRING, "56")],
        ),
        #   ('#(child)|lt|2',[('ROOT', None),]),
        #   ('lt|#(child)|2',[('ROOT', None),]),
        ("lt|field|1", [(op.LT, None), (op.STRING, "field"), (op.INT, 1)]),
        ("lte|field|2", [(op.LTE, None), (op.STRING, "field"), (op.INT, 2)]),
        ("gt|field|3.0", [(op.GT, None),
         (op.STRING, "field"), (op.FLOAT, 3.0)]),
        ("gt|field|hej", [(op.GT, None),
         (op.STRING, "field"), (op.STRING, "hej")]),
        ("gte|field|3.14", [(op.GTE, None),
         (op.STRING, "field"), (op.FLOAT, 3.14)]),
        #    ('range|field|3.14|4.16',[
        #        ('ROOT', None),
        #    ]),
        (
            "or||exists|field||missing|field||and||equals|wf|bov",
            [
                (op.OR, None),
                (op.EXISTS, None),
                (op.STRING, "field"),
                (op.MISSING, None),
                (op.STRING, "field"),
                (op.AND, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "bov"),
            ],
        ),
        (
            "exists||or|field1|field2|field3",
            [
                (op.EXISTS, None),
                (op.ARG_OR, None),
                (op.STRING, "field1"),
                (op.STRING, "field2"),
                (op.STRING, "field3"),
            ],
        ),
        (
            "and||exists|field||and||or||missing|field||contains|field|string||not||contains|field|string",
            [
                (op.AND, None),
                (op.EXISTS, None),
                (op.STRING, "field"),
                (op.AND, None),
                (op.OR, None),
                (op.MISSING, None),
                (op.STRING, "field"),
                (op.CONTAINS, None),
                (op.STRING, "field"),
                (op.STRING, "string"),
                (op.NOT, None),
                (op.CONTAINS, None),
                (op.STRING, "field"),
                (op.STRING, "string"),
            ],
        ),
        (
            "and||not||equals|wf|satt||or||exists|wf||contains|wf|sitta",
            [
                (op.AND, None),
                (op.NOT, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "satt"),
                (op.OR, None),
                (op.EXISTS, None),
                (op.STRING, "wf"),
                (op.CONTAINS, None),
                (op.STRING, "wf"),
                (op.STRING, "sitta"),
            ],
        ),
        (
            "not||and||equals|wf|satt||or||exists|wf||contains|wf|sitta",
            [
                (op.NOT, None),
                (op.AND, None),
                (op.EQUALS, None),
                (op.STRING, "wf"),
                (op.STRING, "satt"),
                (op.OR, None),
                (op.EXISTS, None),
                (op.STRING, "wf"),
                (op.CONTAINS, None),
                (op.STRING, "wf"),
                (op.STRING, "sitta"),
            ],
        ),
        (
            "contains|name||or|vi|bo",
            [
                (op.CONTAINS, None),
                (op.STRING, "name"),
                (op.ARG_OR, None),
                (op.STRING, "vi"),
                (op.STRING, "bo"),
            ],
        ),
        (
            "contains|name||or|vi|bo",
            [
                (op.CONTAINS, None),
                (op.STRING, "name"),
                (op.ARG_OR, None),
                (op.STRING, "vi"),
                (op.STRING, "bo"),
            ],
        ),
        pytest.param(
            "equals||and|area||or|population|density||6312",
            [
                (op.EQUALS, None),
                (op.ARG_AND, None),
                (op.STRING, "area"),
                (op.ARG_OR, None),
                (op.STRING, "population"),
                (op.STRING, "density"),
                (op.INT, 6312),
            ],
            marks=pytest.mark.xfail(
                reason="Current query-dsl syntax can't handle this"
            ),
        ),
    ],
)
def test_karp_tng_parser_success(query, facit):
    r = parser.parse(query)

    assert r is not None
    r.pprint()
    _test_nodes(r, facit)



from sqlalchemy import types
import ulid


class ULIDType(types.TypeDecorator):
    impl = types.CHAR(26)

    cache_ok = True

    python_type = ulid.ULID

    @staticmethod
    def _coerce(value):
        if value and not isinstance(value, ulid.ULID):
            value = ulid.parse(value)
        return value

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if not isinstance(value, ulid.ULID):
            value = self._coerce(value)

        return value.str

    def process_result_value(self, value, dialect):
        return value if value is None else ulid.from_str(value)



from typing import Iterable, TypeVar

from .node import Node

AnyValue = TypeVar("AnyValue", str, int, float)


class Ast:
    def __init__(self, root: Node = None):
        self.root = root

    def __repr__(self):
        return "<Tree root={}>".format(repr(self.root))

    def is_empty(self) -> bool:
        return self.root is None

    def pprint(self):
        print("<Tree root=")
        if not self.is_empty():
            self.root.pprint(1)
        print(">")

    # def validate_arity(self) -> Tuple[bool, str]:
    #     if self.is_empty():
    #         return True, "This tree is empty."
    #
    #     result = []
    #     self.root.validate_arity(result)
    #     if not result:
    #         return True, None
    #     else:
    #         return False, ', '.join(result)

    def gen_stream(self) -> Iterable[Node]:
        if not self.is_empty():
            yield from self.root.gen_stream()



import logging
import typing

from karp.lex.application.queries import EntryViews
from karp.search.application.repositories import IndexEntry
from karp.search.application.transformers import PreProcessor, EntryTransformer


logger = logging.getLogger(__name__)


class GenericPreProcessor(PreProcessor):
    def __init__(
        self,
        entry_transformer: EntryTransformer,
        entry_views: EntryViews,
    ):
        super().__init__()
        self.entry_transformer = entry_transformer
        self.entry_views = entry_views

    def process(
        self,
        resource_id: str,
    ) -> typing.Iterable[IndexEntry]:
        logger.debug(
            "processing entries for resource", extra={"resource_id": resource_id}
        )

        for entry in self.entry_views.all_entries(resource_id):
            logger.debug(
                "processing entry", extra={"entry": entry, "resource_id": resource_id}
            )
            yield self.entry_transformer.transform(resource_id, entry)



from .entry_repos import SqlListEntryRepos, SqlReadOnlyEntryRepoRepositry
from .generic_entries import GenericEntryViews, GenericEntryQuery, GenericGetEntryDiff, GenericGetEntryHistory, GenericGetHistory
from .generic_network import GenericGetReferencedEntries
from .generic_resources import GenericGetEntryRepositoryId
from .resources import SqlGetPublishedResources, SqlGetResources, SqlReadOnlyResourceRepository



import collections
import logging
import typing

import logging

from karp.lex.application.queries.resources import ResourceDto

from karp.lex.domain import entities, errors as lex_errors
from karp.lex.application.queries import (
    GetReferencedEntries,
    ReadOnlyResourceRepository,
    EntryViews,
    EntryDto,
)
from karp.lex.application.repositories import (
    ResourceUnitOfWork,
    EntryUowRepositoryUnitOfWork,
)
from karp.search.application.transformers import EntryTransformer
from karp.search.application.repositories import IndexUnitOfWork, IndexEntry


logger = logging.getLogger(__name__)


class GenericEntryTransformer(EntryTransformer):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        resource_repo: ReadOnlyResourceRepository,
        entry_views: EntryViews,
        get_referenced_entries: GetReferencedEntries,
    ) -> None:
        super().__init__()
        self.index_uow = index_uow
        self.resource_repo = resource_repo
        self.entry_views = entry_views
        self.get_referenced_entries = get_referenced_entries

    def transform(self, resource_id: str, src_entry: EntryDto) -> IndexEntry:
        """
        TODO This is very slow (for resources with references) because of db-lookups everywhere in the code
        TODO We can pre-fetch the needed entries (same code that only looks for refs?) or
        TODO somehow get the needed entries in bulk after transforming some entries and insert them into
        TODO the transformed entries afterward. Very tricky.
        """
        logger.debug(
            "transforming entry",
            extra={"entity_id": src_entry.entity_id, "resource_id": resource_id},
        )
        index_entry = self.index_uow.repo.create_empty_object()
        index_entry.id = str(src_entry.entity_id)
        self.index_uow.repo.assign_field(
            index_entry, "_entry_version", src_entry.version
        )
        self.index_uow.repo.assign_field(
            index_entry, "_last_modified", src_entry.last_modified
        )
        self.index_uow.repo.assign_field(
            index_entry, "_last_modified_by", src_entry.last_modified_by
        )
        resource = self.resource_repo.get_by_resource_id(resource_id)
        if not resource:
            raise lex_errors.ResourceNotFound(None, resource_id=resource_id)
        self._transform_to_index_entry(
            resource,
            # resource_repo,
            # indexer,
            src_entry.entry,
            index_entry,
            resource.config["fields"].items(),
        )
        logger.debug(
            "transformed entry", extra={"entry": src_entry, "index_entry": index_entry}
        )
        return index_entry

    def _transform_to_index_entry(
        self,
        resource: ResourceDto,
        # resource_repo: ResourceRepository,
        # indexer: SearchService,
        _src_entry: typing.Dict,
        _index_entry: IndexEntry,
        fields,
    ):
        logger.debug("transforming [part of] entry", extra={"src_entry": _src_entry})
        for field_name, field_conf in fields:
            field_content = None
            if field_conf.get("virtual"):
                logger.debug("found virtual field")
                if res := self._evaluate_function(
                    field_conf["function"], _src_entry, resource
                ):
                    self.index_uow.repo.assign_field(
                        _index_entry, f"v_{field_name}", res
                    )
            elif field_conf.get("collection"):
                field_content = self.index_uow.repo.create_empty_list()
                if field_name in _src_entry:
                    for subfield in _src_entry[field_name]:
                        if field_conf["type"] == "object":
                            subfield_content = self.index_uow.repo.create_empty_object()
                            self._transform_to_index_entry(
                                resource,
                                subfield,
                                subfield_content,
                                field_conf["fields"].items(),
                            )
                            self.index_uow.repo.add_to_list_field(
                                field_content, subfield_content.entry
                            )
                        else:
                            self.index_uow.repo.add_to_list_field(
                                field_content, subfield
                            )
            elif field_conf["type"] == "object":
                field_content = self.index_uow.repo.create_empty_object()
                if field_name in _src_entry:
                    self._transform_to_index_entry(
                        resource,
                        _src_entry[field_name],
                        field_content,
                        field_conf["fields"].items(),
                    )
            elif field_conf["type"] in (
                "integer",
                "string",
                "number",
                "boolean",
                "long_string",
            ):
                if field_name in _src_entry:
                    field_content = _src_entry[field_name]

            if field_content:
                self.index_uow.repo.assign_field(
                    _index_entry, field_name, field_content
                )

            # Handle ref
            if field_conf.get("ref") and field_name in _src_entry:
                if res := self._resolve_ref(
                    resource, _src_entry, field_conf["ref"], field_name
                ):
                    self.index_uow.repo.assign_field(
                        _index_entry, f"v_{field_name}", res
                    )

    def _resolve_ref(
        self,
        resource: ResourceDto,
        src_entry: dict,
        ref_conf: dict,
        field_name: str,
    ) -> typing.Optional[typing.Any]:
        res = None
        if "resource_id" in ref_conf:
            ref_resource = self.resource_repo.get_by_resource_id(
                ref_conf["resource_id"], version=ref_conf.get("resource_version")
            )
        else:
            ref_resource = resource

        if not ref_resource:
            return res
        try:
            if ref_conf["field"].get("collection"):
                res = self.index_uow.repo.create_empty_list()
                for ref_id in src_entry[field_name]:
                    if ref_entry_body := self.entry_views.get_by_id_optional(
                        ref_resource.resource_id, str(ref_id)
                    ):
                        # ref_entry = {
                        #     field_name: ref_entry_body.entry
                        # }
                        ref_entry = ref_entry_body.entry
                        if ref_conf["field"]["type"] == "object":
                            ref_index_entry = self.index_uow.repo.create_empty_object()
                            for ref_field_name, _ref_field_conf in ref_conf["field"][
                                "fields"
                            ].items():
                                self.index_uow.repo.assign_field(
                                    ref_index_entry,
                                    ref_field_name,
                                    ref_entry[ref_field_name],
                                )
                                self.index_uow.repo.add_to_list_field(
                                    res, ref_index_entry
                                )
                    # ref_objs = []
                    # if ref_resource:
                    #     for ref_id in _src_entry[field_name]:
                    #         ref_entry_body = self.entry_views.get_by_entry_id_optional(
                    #             ref_field['resource_id'], str(ref_id))
                    #         if ref_entry_body:
                    #             ref_entry = {
                    #                 field_name: ref_entry_body.entry}
                    #             ref_index_entry = (
                    #                 self.index_uow.repo.create_empty_object()
                    #             )
                    #             list_of_sub_fields = (
                    #                 (field_name, ref_field["field"]),)
                    #             self._transform_to_index_entry(
                    #                 resource,
                    #                 # resource_repo,
                    #                 # indexer,
                    #                 ref_entry,
                    #                 ref_index_entry,
                    #                 list_of_sub_fields,
                    #             )
                    #             ref_objs.append(
                    #                 ref_index_entry.entry[field_name])
                    # self.index_uow.repo.assign_field(
                    #     _index_entry, "v_" + field_name, ref_objs
                    # )
            elif ref := self.entry_views.get_by_id_optional(
                ref_resource.resource_id, str(src_entry[field_name])
            ):
                ref_entry = {field_name: ref.entry}
                ref_index_entry = self.index_uow.repo.create_empty_object()
                list_of_sub_fields = ((field_name, ref_conf["field"]),)
                self._transform_to_index_entry(
                    ref_resource,
                    ref_entry,
                    ref_index_entry,
                    list_of_sub_fields,
                )
                res = ref_index_entry.entry[field_name]
            return res
        except ValueError as exc:
            logger.warning(
                "Got ValueError (%s)",
                str(exc),
                extra={"exc": exc, "resource": resource},
            )

        # if field_conf["type"] == "object":
        #     logger.debug("found field with type 'object'")
        #     field_content = self.index_uow.repo.create_empty_object()
        #     if field_name in _src_entry:
        #         self._transform_to_index_entry(
        #             resource,
        #             # resource_repo,
        #             # indexer,
        #             _src_entry[field_name],
        #             field_content,
        #             field_conf["fields"].items(),
        #         )
        # else:
        #     field_content = _src_entry.get(field_name)

    def _evaluate_function(
        self,
        # resource_repo: ResourceRepository,
        # indexer: SearchService,
        function_conf: typing.Dict,
        src_entry: typing.Dict,
        src_resource: ResourceDto,
    ):
        logger.debug(
            "indexing._evaluate_function",
            extra={"src_resource": src_resource.resource_id},
        )
        logger.debug("indexing._evaluate_function", extra={"src_entry": src_entry})
        if "multi_ref" in function_conf:
            function_conf = function_conf["multi_ref"]
            target_field = function_conf["field"]
            if "resource_id" in function_conf:
                logger.debug(
                    "indexing._evaluate_function: trying to find '%s'",
                    function_conf["resource_id"],
                )
                target_resource = self.resource_repo.get_by_resource_id(
                    function_conf["resource_id"],
                    version=function_conf["resource_version"],
                )
                if target_resource is None:
                    logger.warning(
                        "Didn't find the resource with resource_id='%s'",
                        function_conf["resource_id"],
                    )
                    return self.index_uow.repo.create_empty_list()
            else:
                target_resource = src_resource
            logger.debug(
                "indexing._evaluate_function target_resource='%s'",
                target_resource.resource_id,
            )
            if "test" in function_conf:
                operator, args = list(function_conf["test"].items())[0]
                if operator in ["equals", "contains"]:
                    filters = {"discarded": False}
                    for arg in args:
                        if "self" in arg:
                            filters[target_field] = src_entry[arg["self"]]
                        else:
                            raise NotImplementedError()
                    # target_entries = entryread.get_entries_by_column(
                    #     target_resource, filters
                    # )
                    target_entries = self.entry_views.get_by_referenceable(
                        target_resource.resource_id, filters
                    )
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

            res = self.index_uow.repo.create_empty_list()
            for entry in target_entries:
                index_entry = self.index_uow.repo.create_empty_object()
                list_of_sub_fields = (("tmp", function_conf["result"]),)
                self._transform_to_index_entry(
                    target_resource,
                    # resource_repo,
                    # indexer,
                    {"tmp": entry.entry},
                    index_entry,
                    list_of_sub_fields,
                )
                self.index_uow.repo.add_to_list_field(res, index_entry.entry["tmp"])
        elif "plugin" in function_conf:
            plugin_id = function_conf["plugin"]
            import karp.pluginmanager as plugins

            res = plugins.plugins[plugin_id].apply_plugin_function(
                src_resource.entity_id, src_resource.version, src_entry
            )
        else:
            raise NotImplementedError()
        return res

    def update_references(
        self,
        resource_id: str,
        # resource_repo: ResourceRepository,
        # indexer: SearchService,
        # resource: entities.Resource,
        entry_ids: typing.Iterable[str],
    ) -> None:
        add = collections.defaultdict(list)
        for src_entry_id in entry_ids:
            refs = self.get_referenced_entries.query(resource_id, src_entry_id)
            for field_ref in refs:
                ref_resource = self.resource_repo.get_by_resource_id(
                    field_ref.resource_id, version=(field_ref.resource_version)
                )
                if ref_resource:
                    ref_index_entry = self.transform(
                        # resource_repo,
                        # indexer,
                        ref_resource.resource_id,  # TODO use resource directly
                        field_ref.entry,
                        # ref_resource.config["fields"].items(),
                    )
                    # metadata = resourcemgr.get_metadata(ref_resource, field_ref["id"])
                    add[field_ref.resource_id].append(ref_index_entry)

        for ref_resource_id, ref_entries in add.items():
            self.index_uow.repo.add_entries(ref_resource_id, ref_entries)



import typing
from karp.foundation.value_objects import unique_id
from karp.lex.application.repositories import ResourceUnitOfWork
from karp.search.application.queries import ResourceViews


class GenericResourceViews(ResourceViews):
    def __init__(self, resource_uow: ResourceUnitOfWork) -> None:
        super().__init__()
        self._resource_uow = resource_uow

    def get_resource_config(self, resource_id: str) -> typing.Dict:
        with self._resource_uow as uw:
            return uw.repo.by_resource_id(resource_id).config

    def get_resource_ids(self, repo_id: unique_id.UniqueId) -> typing.List[str]:
        with self._resource_uow as uw:
            return [
                resource.resource_id
                for resource in uw.repo.get_all_resources()
                if resource.entry_repository_id == repo_id
            ]



import enum
import logging
import typing
from typing import Dict, Optional

import regex
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session


DUPLICATE_PROG = regex.compile(r"Duplicate entry '(.+)' for key '(\w+)'")

_create_new = object()

logger = logging.getLogger(__name__)


class SqlUnitOfWork:  # (repositories.UnitOfWork):
    class State(enum.Enum):
        initialized = 0
        begun = 1
        committed = 2
        aborted = 3

    def __init__(
        self,
        *,
        session: Optional[Session] = None,
    ):
        logger.debug('Init sqlunitofwork session=%s', session)
        self._session = session
        self._session_is_created_here = self._session is None
        self._state = SqlUnitOfWork.State.initialized

    # @property
    # def repo(self):
    #     return self._repo

    def __enter__(self):
        return self.begin()

    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     if exc_type is None:
    #         # self.commit()
    #         if self._session_is_created_here:
    #             self.close()
    #         return False
    #     self.abort()
    #     if self._session_is_created_here:
    #         self.close()
    #     return False  # re-raise

    def begin(self):
        self._check_state(expected_state=SqlUnitOfWork.State.initialized)
        self._state = SqlUnitOfWork.State.begun
        return self._begin()

    def _begin(self):
        return self

    def _check_state(self, expected_state):
        if self._state != expected_state:
            pass
            # logger.warning(
            #     "State conflict. repositories is in state '%s' and not '%s'",
            #     self._state,
            #     expected_state,
            # )
            # raise RuntimeError(
            #     f"State conflict. repositories is in state '{self._state!s}' and not '{expected_state!s}'"
            # )

    def _commit(self):
        self._check_state(expected_state=SqlUnitOfWork.State.begun)
        # try:
        logger.info('About to commit', extra={
                    'session': self._session, 'session_new': self._session.new if self._session else None})
        self._session.commit()
        logger.info('commited', extra={
                    'session': self._session, 'session_new': self._session.new if self._session else None})

        # self._state = SqlUnitOfWork.State.initialized
        # except db.exc.IntegrityError as err:
        #     logger.exception(err)
        #     str_err = str(err)
        #     print(f"str(err) = {str_err}")
        #     if "Duplicate entry" in str_err:
        #         match = DUPLICATE_PROG.search(str_err)
        #         if match:
        #             value = match.group(1)
        #             key = match.group(2)
        #             if key == "PRIMARY":
        #                 key = self.primary_key()
        #         else:
        #             value = "UNKNOWN"
        #             key = "UNKNOWN"
        #         raise errors.IntegrityError(key=key, value=value) from err
        #     raise errors.IntegrityError("Unknown integrity error") from err

    def abort(self):
        self._check_state(expected_state=SqlUnitOfWork.State.begun)
        self._session.rollback()
        self._state = SqlUnitOfWork.State.initialized

    def rollback(self):
        logger.debug('rollback called')
        self._check_state(expected_state=SqlUnitOfWork.State.begun)
        self._session.rollback()
        self._state = SqlUnitOfWork.State.initialized

    def _close(self):
        if self._session_is_created_here:
            logger.debug('closing session=%s', self._session)
            self._session.close()



import json
import logging
import typing
from pathlib import Path
from typing import IO, Dict, Generic, List, Optional, Tuple

import logging

from karp.lex.domain import errors, entities
from karp.lex.domain.entities import Resource
from karp.foundation import events as foundation_events
from karp.foundation.commands import CommandHandler
from karp.lex.application.queries import ResourceDto
from karp.lex.application import repositories as lex_repositories
from karp.lex.domain import commands
from karp.lex.application import repositories


logger = logging.getLogger(__name__)

# resource_models = {}  # Dict
# history_models = {}  # Dict
# resource_configs = {}  # Dict
# resource_versions = {}  # Dict[str, int]
# field_translations = {}  # Dict[str, Dict[str, List[str]]]


class BasingResource:
    def __init__(self, resource_uow: repositories.ResourceUnitOfWork) -> None:
        self.resource_uow = resource_uow

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


class CreatingResource(CommandHandler[commands.CreateResource], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        entry_repo_uow: lex_repositories.EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__(resource_uow=resource_uow)
        self.entry_repo_uow = entry_repo_uow

    def execute(self, command: commands.CreateResource) -> ResourceDto:
        with self.entry_repo_uow as uow:
            entry_repo_exists = uow.repo.get_by_id_optional(command.entry_repo_id)
            if not entry_repo_exists:
                raise errors.NoSuchEntryRepository(
                    f"Entry repository '{command.entry_repo_id}' not found"
                )

        with self.resource_uow as uow:
            existing_resource = uow.repo.get_by_resource_id_optional(
                command.resource_id
            )
            if (
                existing_resource
                and not existing_resource.discarded
                and existing_resource.entity_id != command.entity_id
            ):
                raise errors.IntegrityError(
                    f"Resource with resource_id='{command.resource_id}' already exists."
                )

            resource = entities.create_resource(
                entity_id=command.entity_id,
                resource_id=command.resource_id,
                config=command.config,
                message=command.message,
                entry_repo_id=command.entry_repo_id,
                created_at=command.timestamp,
                created_by=command.user,
                name=command.name,
            )

            uow.repo.save(resource)
            uow.commit()
        return ResourceDto(**resource.dict())

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


class SettingEntryRepoId(CommandHandler[commands.SetEntryRepoId], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        entry_repo_uow: lex_repositories.EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__(resource_uow=resource_uow)
        self.entry_repo_uow = entry_repo_uow

    def execute(self, command: commands.SetEntryRepoId) -> None:
        with self.entry_repo_uow as uow:
            entry_repo_exists = uow.repo.get_by_id_optional(command.entry_repo_id)
            if not entry_repo_exists:
                raise errors.NoSuchEntryRepository(
                    f"Entry repository '{command.entry_repo_id}' not found"
                )

        with self.resource_uow as uow:
            resource = uow.repo.get_by_resource_id(command.resource_id)

            resource.set_entry_repo_id(
                entry_repo_id=command.entry_repo_id,
                user=command.user,
                timestamp=command.timestamp,
            )

            uow.repo.save(resource)
            uow.commit()


class UpdatingResource(CommandHandler[commands.UpdateResource], BasingResource):
    def __init__(self, resource_uow: repositories.ResourceUnitOfWork) -> None:
        super().__init__(resource_uow=resource_uow)

    def execute(self, command: commands.UpdateResource):
        with self.resource_uow as uow:
            resource = uow.repo.by_resource_id(command.resource_id)
            if resource.update(
                name=command.name,
                config=command.config,
                user=command.user,
                message=command.message,
                timestamp=command.timestamp,
                version=command.version,
            ):

                uow.repo.save(resource)
            uow.commit()

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


class PublishingResource(CommandHandler[commands.PublishResource], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        **kwargs,
    ) -> None:
        super().__init__(resource_uow=resource_uow)

    def execute(self, command: commands.PublishResource):
        logger.info("publishing resource", extra={"resource_id": command.resource_id})
        with self.resource_uow as uow:
            resource = uow.repo.by_resource_id(command.resource_id)
            if not resource:
                raise errors.ResourceNotFound(command.resource_id)
            resource.publish(
                user=command.user,
                message=command.message,
                timestamp=command.timestamp,
                version=command.version,
            )
            uow.repo.save(resource)
            uow.commit()

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()


class DeletingResource(CommandHandler[commands.DeleteResource], BasingResource):
    def __init__(
        self,
        resource_uow: repositories.ResourceUnitOfWork,
        **kwargs,
    ) -> None:
        super().__init__(resource_uow=resource_uow)

    def execute(self, command: commands.DeleteResource):
        logger.info("deleting resource", extra={"resource_id": command.resource_id})
        with self.resource_uow as uow:
            resource = uow.repo.by_resource_id(command.resource_id)
            if not resource:
                raise errors.ResourceNotFound(command.resource_id)
            resource.discard(
                user=command.user, message=command.message, timestamp=command.timestamp
            )
            uow.repo.save(resource)
            uow.commit()

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()



import abc

from karp import lex
from karp.search.application.repositories import IndexEntry


class EntryTransformer(abc.ABC):
    @abc.abstractmethod
    def transform(
        self,
        resource_id: str,
        src_entry: lex.EntryDto,
    ) -> IndexEntry:
        pass



import elasticsearch
import injector
from sqlalchemy.orm import sessionmaker
import logging

from karp.foundation.events import EventBus
from karp.lex.application.queries import (
    GetReferencedEntries,
    GetEntryRepositoryId,
    ReadOnlyResourceRepository,
    EntryViews,
)
from karp.lex.application.repositories import (
    ResourceUnitOfWork,
    EntryUowRepositoryUnitOfWork,
)
from karp import lex, search
from karp.search.application.queries import (
    ResourceViews,
    SearchService,
)
from karp.search.application.repositories import IndexUnitOfWork

from karp.search.application.transformers import (
    EntryTransformer,
    PreProcessor,
)
from karp.search_infrastructure.queries import (
    GenericPreviewEntry,
    GenericResourceViews,
    GenericSearchService,
    Es6SearchService,
)
from karp.search_infrastructure.transformers import (
    GenericEntryTransformer,
    GenericPreProcessor,
)
from karp.search_infrastructure.repositories import (
    # SqlIndexUnitOfWork,
    NoOpIndexUnitOfWork,
    Es6IndexUnitOfWork,
)


logger = logging.getLogger(__name__)


class SearchInfrastructure(injector.Module):
    @injector.provider
    def entry_transformer(
        self,
        index_uow: IndexUnitOfWork,
        resource_repo: lex.ReadOnlyResourceRepository,
        entry_views: EntryViews,
        get_referenced_entries: GetReferencedEntries,
    ) -> EntryTransformer:
        return GenericEntryTransformer(
            index_uow=index_uow,
            resource_repo=resource_repo,
            entry_views=entry_views,
            get_referenced_entries=get_referenced_entries,
        )

    @injector.provider
    def pre_processor(
        self,
        entry_transformer: EntryTransformer,
        entry_views: EntryViews,
    ) -> PreProcessor:
        return GenericPreProcessor(
            entry_transformer=entry_transformer,
            entry_views=entry_views,
        )

    @injector.provider
    def preview_entry(
        self,
        entry_transformer: EntryTransformer,
        resource_uow: lex.ResourceUnitOfWork,
    ) -> search.PreviewEntry:
        return GenericPreviewEntry(
            entry_transformer=entry_transformer,
            resource_uow=resource_uow,
        )


class GenericSearchInfrastructure(injector.Module):
    @injector.provider
    def get_resource_config(self, resource_uow: lex.ResourceUnitOfWork) -> ResourceViews:
        return GenericResourceViews(
            resource_uow=resource_uow
        )


class GenericSearchIndexMod(injector.Module):

    @injector.provider
    def generic_search_service(
        self,
        get_entry_repo_id: GetEntryRepositoryId,
        entry_uow_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> SearchService:
        return GenericSearchService(
            get_entry_repo_id=get_entry_repo_id,
            entry_uow_repo_uow=entry_uow_repo_uow,
        )

    @injector.provider
    def noop_index_uow(
        self,
    ) -> IndexUnitOfWork:
        return NoOpIndexUnitOfWork()


class Es6SearchIndexMod(injector.Module):

    @injector.provider
    def es6_search_service(
        self,
        es: elasticsearch.Elasticsearch,
    ) -> SearchService:
        return Es6SearchService(es=es)

    @injector.provider
    def es6_index_uow(
        self,
        es: elasticsearch.Elasticsearch,
        event_bus: EventBus,
    ) -> IndexUnitOfWork:
        return Es6IndexUnitOfWork(
            es=es,
            event_bus=event_bus
        )



from typing import Callable, Type
from fastapi import Depends
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
from starlette.requests import Request

from karp import lex
from karp.foundation.events import EventBus
from karp.lex import (
    EntryRepositoryUnitOfWorkFactory,
    EntryUowRepositoryUnitOfWork,
    ResourceUnitOfWork,
)
from karp.webapp.dependencies import db_deps
from karp.webapp.dependencies.db_deps import get_database, get_session
from karp.webapp.dependencies import event_deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req

from karp.db_infrastructure import Database

from karp.lex_infrastructure import (
    GenericGetEntryDiff,
    GenericGetEntryHistory,
    GenericGetHistory,
    SqlEntryUowRepositoryUnitOfWork,
    SqlGetPublishedResources,
    SqlResourceUnitOfWork,
    SqlReadOnlyResourceRepository,
)
from karp.lex_infrastructure.repositories import SqlResourceRepository


def get_resource_repository(db_session: Session = Depends(get_session)) -> SqlResourceRepository:
    return SqlResourceRepository(db_session)


def get_resource_unit_of_work(
    db_session: Session = Depends(get_session),
    event_bus: EventBus = Depends(event_deps.get_eventbus),
) -> ResourceUnitOfWork:
    return SqlResourceUnitOfWork(
        event_bus=event_bus,
        session=db_session,
    )

def get_entry_repo_uow(
    db_session: Session = Depends(get_session),
    entry_uow_factory: EntryRepositoryUnitOfWorkFactory = Depends(inject_from_req(EntryRepositoryUnitOfWorkFactory)),
    event_bus: EventBus = Depends(event_deps.get_eventbus),
) -> EntryUowRepositoryUnitOfWork:
    return SqlEntryUowRepositoryUnitOfWork(
        event_bus=event_bus,
        entry_uow_factory=entry_uow_factory,
        session=db_session,
    )


def get_lex_uc(Use_case_type: Type) -> Callable:
    def factory(
        resource_uow: ResourceUnitOfWork = Depends(get_resource_unit_of_work),
        entry_repo_uow: EntryUowRepositoryUnitOfWork = Depends(get_entry_repo_uow),
    ) -> Use_case_type:
        return Use_case_type(
            resource_uow=resource_uow,
            entry_repo_uow=entry_repo_uow,
        )
    return factory


def get_resources_read_repo(
    conn: Connection = Depends(db_deps.get_connection)
) -> lex.ReadOnlyResourceRepository:
    return SqlReadOnlyResourceRepository(conn)


def get_published_resources(
    conn: Connection = Depends(db_deps.get_connection)
) -> lex.GetPublishedResources:
    return SqlGetPublishedResources(conn)


def get_entry_diff(
    resource_uow: ResourceUnitOfWork = Depends(get_resource_unit_of_work),
    entry_repo_uow: EntryUowRepositoryUnitOfWork = Depends(get_entry_repo_uow),
) -> lex.GetEntryDiff:
    return GenericGetEntryDiff(
        resource_uow=resource_uow,
        entry_repo_uow=entry_repo_uow,
    )


def get_entry_history(
    resource_uow: ResourceUnitOfWork = Depends(get_resource_unit_of_work),
    entry_repo_uow: EntryUowRepositoryUnitOfWork = Depends(get_entry_repo_uow),
) -> lex.GetEntryHistory:
    return GenericGetEntryHistory(
        resource_uow=resource_uow,
        entry_repo_uow=entry_repo_uow,
    )


def get_history(
    resource_uow: ResourceUnitOfWork = Depends(get_resource_unit_of_work),
    entry_repo_uow: EntryUowRepositoryUnitOfWork = Depends(get_entry_repo_uow),
) -> lex.GetHistory:
    return GenericGetHistory(
        resource_uow=resource_uow,
        entry_repo_uow=entry_repo_uow,
    )



"""Lex domain errors."""


from typing import Dict
from karp.foundation.errors import NotFoundError


class LexDomainError(Exception):

    """Base exception for domain errors."""

    def __init__(self, *args: object, **kwargs) -> None:
        super().__init__(*args)
        self.extras = kwargs or {}


class ConsistencyError(LexDomainError):
    pass


class ConstraintsError(LexDomainError):
    pass


class DiscardedEntityError(LexDomainError):
    pass


class EntryNotFound(NotFoundError, LexDomainError):
    entity_name = "Entry"

    def __init__(self, *args, entity_id=None, **kwargs):
        NotFoundError.__init__(self, entity_id, *args)
        LexDomainError.__init__(self, **kwargs)
        # super().__init__(entity_id, *args, **kwargs)


class EntryRepoNotFound(NotFoundError, LexDomainError):
    entity_name = "EntryRepository"

    def __init__(self, entity_id, *args, **kwargs):
        NotFoundError.__init__(self, entity_id, *args)
        LexDomainError.__init__(self, entity_id, *args, **kwargs)


class ResourceNotFound(NotFoundError, LexDomainError):
    entity_name = "Resource"

    def __init__(self, entity_id, *args, **kwargs):
        NotFoundError.__init__(self, entity_id, *args)
        LexDomainError.__init__(self, entity_id, *args, **kwargs)


class IntegrityError(LexDomainError):
    pass


class RepositoryError(LexDomainError):
    pass


class RepositoryStatusError(LexDomainError):
    pass


class NonExistingField(RepositoryError):
    """Raised when a field doesn't exist in a repo."""

    def __init__(self, field: str, **kwargs):
        # if "code" not in kwargs:
        #     kwargs["code"] = errors.ClientErrorCodes.DB_GENERAL_ERROR
        super().__init__(f"Field '{field}' doesn't exist.", **kwargs)


class InvalidEntry(LexDomainError):
    pass


class MissingIdField(InvalidEntry):
    def __init__(self, resource_id: str, entry: Dict) -> None:
        super().__init__(f"Missing ID field for resource '{resource_id}' in '{entry}'")
        self.error_obj = {
            "loc": {"resource_id": resource_id},
            "data": {"entry": entry},
            "error": {"message": str(self), "code": "missing id field"},
        }


class UpdateConflict(LexDomainError):
    def __init__(self, diff):
        super().__init__(
            "Version conflict. Please update entry.",
        )
        self.error_obj = {"diff": diff, "error": str(self)}


class LexValueError(ValueError, LexDomainError):
    pass


class InvalidResourceId(LexValueError):
    pass


class LogicError(LexDomainError):
    pass


class InvalidEntrySchema(ValueError, LogicError):
    pass


class NoSuchEntryRepository(ValueError, LexDomainError):
    pass


class DiffImposible(LexDomainError):
    pass



"""Tests for SQLResourceRepository"""

import pytest

from karp.lex.domain.errors import IntegrityError
from karp.lex.domain.entities.resource import Resource, ResourceOp
from karp.lex_infrastructure.repositories.sql_resources import SqlResourceRepository
from karp.tests.unit.lex import factories


@pytest.fixture(name="resource_repo")  # , scope="module")
def fixture_resource_repo(sqlite_session_factory):
    session = sqlite_session_factory()
    return SqlResourceRepository(session)


def test_sql_resource_repo_empty(resource_repo):
    assert resource_repo.resource_ids() == []
    assert resource_repo.history_by_resource_id("test_id") == []


def test_sql_resource_repo_put_resource(resource_repo):
    resource = factories.ResourceFactory()

    resource_repo.save(resource)
    expected_version = 1

    assert resource_repo.resource_ids() == [resource.resource_id]

    resource_copy_1 = resource_repo.by_id(resource.id)
    assert resource_copy_1.version == resource.version
    assert resource_copy_1.message == resource.message
    assert resource_copy_1.op == ResourceOp.ADDED

    resource_id_history = resource_repo.history_by_resource_id(resource.resource_id)
    assert len(resource_id_history) == 1

    test_lex = resource_repo.by_resource_id(resource.resource_id)

    assert isinstance(test_lex, Resource)
    assert isinstance(test_lex.config, dict)
    assert test_lex.resource_id == resource.resource_id
    assert test_lex.id == resource.id
    assert test_lex.name == resource.name
    assert test_lex.version == resource.version


#     # Update resource
#     with unit_of_work(using=resource_repo) as uw:
#         resource.config["c"] = "added"
#         resource.config["a"] = "changed"
#         resource.is_published = True
#         resource.stamp(user="Test user", message="change config")
#         uw.update(resource)
#         assert resource.version == 2
#
#     with unit_of_work(using=resource_repo) as uw:
#         test_lex = uw.by_resource_id(resource_id)
#
#         assert test_lex is not None
#         assert test_lex.config["a"] == "changed"
#         assert test_lex.config["c"] == "added"
#         assert test_lex.is_published is True
#         assert test_lex.version == 2
#         assert uw.get_latest_version(resource_id) == test_lex.version
#
#     # Test history
#     with unit_of_work(using=resource_repo) as uw:
#         resource_id_history = uw.history_by_resource_id(resource_id)
#
#         assert len(resource_id_history) == 2
#         assert resource_id_history[0].version == 2
#         assert resource_id_history[1].version == 1


# def test_sql_resource_repo_put_lexical_resource(resource_repo):
#     resource_id = "test_id"
#     resource_name = "Test"
#     resource_config = {
#         "resource_id": resource_id,
#         "resource_name": resource_name,
#         "a": "b",
#     }
#     resource = Resource.create_resource("lexical_resource", resource_config)

#     expected_version = 1

#     with unit_of_work(using=resource_repo) as uw:
#         uw.save(resource)
#         uw.commit()
#         assert uw.resource_ids() == [resource_id]

#         assert resource.version == expected_version
#         assert resource.message == "Resource added."
#         assert resource.op == ResourceOp.ADDED
#         resource_id_history = uw.history_by_resource_id(resource_id)
#         assert len(resource_id_history) == 1

#     with unit_of_work(using=resource_repo) as uw:
#         test_lex = uw.resource_with_id_and_version(resource_id, expected_version)

#         assert isinstance(test_lex, Resource)
#         assert isinstance(test_lex, LexicalResource)
#         assert isinstance(test_lex.config, dict)
#         assert test_lex.resource_id == resource_id
#         assert test_lex.id == resource.id
#         assert test_lex.name == resource_name


#     {"resource_id": resource_id, "resource_name": resource_name, "a": "b"}
# )


def test_sql_resource_repo_update_resource(resource_repo):
    resource = factories.ResourceFactory(
        config={"a": "b"},
    )
    resource_repo.save(resource)

    resource.set_config(
        config={"a": "changed", "c": "added"},
        message="change config",
        user="Test user",
        version=1,
        # timestamp=time.utc_now(),
    )
    resource_repo.save(resource)

    assert resource_repo.by_id(resource.id).version == 2
    assert resource_repo.by_resource_id(resource.resource_id).version == 2

    assert resource_repo.get_by_id(resource.entity_id, version=1).version == 1
    assert (
        resource_repo.get_by_resource_id(resource.resource_id, version=1).version == 1
    )

    lex = resource_repo.by_resource_id(resource.resource_id)
    assert lex is not None
    assert lex.resource_id == resource.resource_id
    assert resource_repo.get_latest_version(resource.resource_id) == resource.version

    lex.set_config(
        config={"a": "changed", "c": "added", "d": "added"},
        user="kristoff@example.com",
        version=lex.version,
        # message='update'
    )
    assert lex.version == 3
    resource_repo.save(lex)

    assert resource_repo.get_by_resource_id(resource.resource_id).version == 3


def test_sql_resource_repo_put_another_resource(resource_repo):
    resource = factories.ResourceFactory()
    resource_repo.save(resource)

    resource2 = factories.ResourceFactory()
    resource_repo.save(resource2)

    assert resource_repo.resource_ids() == [resource.resource_id, resource2.resource_id]


class TestSqlResourceRepo:
    def test_discard_resource_and_insert_new(self, resource_repo):
        resource = factories.ResourceFactory()
        resource_repo.save(resource)

        resource.discard(
            user="kristoff@example.com",
            message="delete",
        )
        resource_repo.save(resource)

        resource2 = factories.ResourceFactory(
            resource_id=resource.resource_id,
        )
        resource_repo.save(resource2)

        assert resource_repo.get_by_resource_id(resource.resource_id).version == 1

        assert resource_repo.resource_ids() == [resource.resource_id]

    def test_change_resource_id_changes_resource_ids(self, resource_repo):
        resource = factories.ResourceFactory()
        resource_repo.save(resource)

        resource.set_resource_id(
            resource_id="changed_id",
            user="kristoff@example.com",
            version=1,
        )
        resource_repo.save(resource)

        assert resource_repo.resource_ids() == ["changed_id"]


# def test_sql_resource_repo_deep_update_of_resource(resource_repo):
#     with unit_of_work(using=resource_repo) as uw:
#         resource = uw.get_active_resource("test_id_2")
#         assert resource is not None

#         resource.config["fields"]["count"] = {"type": "int"}
#         resource.stamp(user="Admin", message="change")
#         assert resource.is_published
#         assert resource.version == 2
#         uw.update(resource)
#         # assert resource.name_id == "test_id_2Test 2"

#     with unit_of_work(using=resource_repo) as uw:
#         resource = uw.get_active_resource("test_id_2")

#         assert resource is not None
#         assert "count" in resource.config["fields"]
#         assert resource.config["fields"]["count"] == {"type": "int"}


# def test_get_published_resources(resource_repo):
#     with unit_of_work(using=resource_repo) as uw:
#         uw.save(create_resource({"resource_id": "test_id_3", "resource_name": "g"}))
#         published_resources = uw.get_published_resources()

#         assert len(published_resources) == 1
#         assert published_resources[0].resource_id == "test_id_2"


# def test_sql_resource_repo_



from typing import Dict, Optional

import pydantic

from karp.foundation.value_objects import UniqueId, make_unique_id

from .base import Command


class CreateEntryRepository(Command):
    entity_id: UniqueId = pydantic.Field(default_factory=make_unique_id)
    repository_type: str
    name: str
    connection_str: Optional[str] = None
    config: Dict
    message: str
    user: str

    @classmethod
    def from_dict(
        cls,
        data: Dict,
        *,
        user: str,
        message: Optional[str] = None
    ):
        return cls(
            entity_id=make_unique_id(),
            repository_type=data.pop('repository_type', 'default'),
            name=data.pop('resource_id'),
            connection_str=data.pop('connection_str', None),
            config=data,
            user=user,
            message=message or 'Entry repository created'
        )


class DeleteEntryRepository(Command):
    entity_id: UniqueId
    message: str
    user: str



import logging

from karp.search.application.repositories import (
    Index, IndexUnitOfWork,
)


logger = logging.getLogger('karp')


class NoOpIndex(Index):
    pass
    def create_index(self, *args, **kwargs):
        logger.debug('create_index called with args=%s, kwargs=%s', args, kwargs)

    def publish_index(self, *args, **kwargs):
        logger.debug('publish_index called with args=%s, kwargs=%s', args, kwargs)

    def add_entries(self, *args, **kwargs):
        logger.debug('add_entries called with args=%s, kwargs=%s', args, kwargs)

    def delete_entry(self, *args, **kwargs):
        logger.debug('delete_entry called with args=%s, kwargs=%s', args, kwargs)


class NoOpIndexUnitOfWork(IndexUnitOfWork):
    def __init__(self):
        self._repo = NoOpIndex()

    @property
    def repo(self):
        return self._repo

    def _commit(self, *args, **kwargs):
        logger.debug('_commit called with args=%s, kwargs=%s', args, kwargs)

    def _close(self, *args, **kwargs):
        logger.debug('_close called with args=%s, kwargs=%s', args, kwargs)

    def rollback(self, *args, **kwargs):
        logger.debug('rollback called with args=%s, kwargs=%s', args, kwargs)



import logging
import typing
from pathlib import Path

import json_streams
from sb_json_tools import jsondiff
import logging

from karp.foundation.commands import CommandHandler
from karp.lex.application.repositories import EntryUnitOfWork
from karp.lex.domain import errors
from karp.lex.domain.entities.entry import Entry
from karp.foundation.value_objects import unique_id

from karp.foundation import events as foundation_events
from karp.lex.domain import commands
from karp.lex.domain.value_objects import EntrySchema
from karp.lex.application import repositories


logger = logging.getLogger(__name__)


class BasingEntry:
    def __init__(
        self,
        entry_repo_uow: repositories.EntryUowRepositoryUnitOfWork,
        resource_uow: repositories.ResourceUnitOfWork,
    ) -> None:
        super().__init__()
        self.entry_repo_uow = entry_repo_uow
        self.resource_uow = resource_uow

    def collect_new_events(self) -> typing.Iterable[foundation_events.Event]:
        yield from self.resource_uow.collect_new_events()
        yield from self.entry_repo_uow.collect_new_events()

    def get_entry_uow(self, entry_repo_id: unique_id.UniqueId) -> EntryUnitOfWork:
        with self.entry_repo_uow as uw:
            return uw.repo.get_by_id(entry_repo_id)


class AddingEntry(BasingEntry, CommandHandler[commands.AddEntry]):
    def execute(self, command: commands.AddEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(command.resource_id)

        # try:
        #     entry_id = resource.id_getter()(command.entry)
        # except KeyError as err:
        #     raise errors.MissingIdField(
        #         resource_id=command.resource_id, entry=command.entry
        #     ) from err
        entry_schema = EntrySchema(resource.entry_json_schema)

        with self.get_entry_uow(resource.entry_repository_id) as uw:
            existing_entry = uw.repo.get_by_id_optional(command.entity_id)
            if (
                existing_entry
                and not existing_entry.discarded
                and existing_entry.entity_id != command.entity_id
            ):
                raise errors.IntegrityError(
                    f"An entry with entry_id '{command.entity_id}' already exists."
                )

            entry_schema.validate_entry(command.entry)

            entry = resource.create_entry_from_dict(
                command.entry,
                user=command.user,
                message=command.message,
                entity_id=command.entity_id,
                timestamp=command.timestamp,
            )
            uw.entries.save(entry)
            uw.commit()
        return entry


class UpdatingEntry(BasingEntry, CommandHandler[commands.UpdateEntry]):
    def execute(self, command: commands.UpdateEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(command.resource_id)

        entry_schema = EntrySchema(resource.entry_json_schema)
        entry_schema.validate_entry(command.entry)

        with self.get_entry_uow(resource.entry_repository_id) as uw:
            try:
                current_db_entry = uw.repo.by_id(command.entity_id)
            except errors.EntryNotFound as err:
                raise errors.EntryNotFound(
                    command.resource_id,
                    entity_id=command.entity_id,
                ) from err

            diff = jsondiff.compare(current_db_entry.body, command.entry)
            if not diff:
                return current_db_entry

            #     db_entry_json = json.dumps(entry)
            #     db_id = current_db_entry.id
            #     latest_history_entry = (
            #         resource.history_model.query.filter_by(entry_id=db_id)
            #         .order_by(resource.history_model.version.desc())
            #         .first()
            #     )
            if not command.force and current_db_entry.version != command.version:
                logger.info(
                    "version conflict",
                    extra={
                        "current_version": current_db_entry.version,
                        "version": command.version,
                    },
                )
                raise errors.UpdateConflict(diff)

            # id_getter = resource.id_getter()
            # new_entry_id = id_getter(command.entry)
            current_db_entry.update_body(
                command.entry,
                user=command.user,
                message=command.message,
                timestamp=command.timestamp,
            )
            # current_db_entry.stamp(
            #     command.user, message=command.message, timestamp=command.timestamp
            # )
            # if new_entry_id != command.entry_id:
            #     logger.info(
            #         "updating entry_id",
            #         extra={
            #             "entry_id": command.entry_id,
            #             "new_entry_id": new_entry_id,
            #         },
            #     )
            #     current_db_entry.entry_id = new_entry_id
            #     # uw.repo.move(current_db_entry, old_entry_id=command.entry_id)
            #     uw.repo.save(current_db_entry)
            # else:
            uw.repo.save(current_db_entry)
            uw.commit()

        return current_db_entry


# def update_entries(*args, **kwargs):
#     return []


# def add_entries_from_file(
#     resource_id: str, version: int, filename: Path
# ) -> list[Entry]:
#     return add_entries(
#         resource_id,
#         json_streams.load_from_file(filename),
#         user_id="local_admin",
#         resource_version=version,
#     )


class AddingEntries(BasingEntry, CommandHandler[commands.AddEntries]):
    def execute(self, command: commands.AddEntries):
        """
        Add entries to DB and INDEX (if present and resource is active).

        Raises
        ------
        RuntimeError
            If the resource.entry_json_schema fails to compile.
        KarpError
            - If an entry fails to be validated against the json schema.
            - If the DB interaction fails.

        Returns
        -------
        List
            List of the id's of the created entries.
        """

        if not isinstance(command.resource_id, str):
            raise ValueError(
                f"'resource_id' must be of type 'str', were '{type(command.resource_id)}'"
            )
        with self.resource_uow:
            resource = self.resource_uow.resources.by_resource_id(command.resource_id)

        if not resource:
            raise errors.ResourceNotFound(command.resource_id)

        entry_schema = EntrySchema(resource.entry_json_schema)

        created_db_entries = []
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            resource.entry_repository_id
        ) as uw:
            for i, entry_raw in enumerate(command.entries):

                entry_schema.validate_entry(entry_raw)

                entry = resource.create_entry_from_dict(
                    entry_raw,
                    user=command.user,
                    message=command.message,
                    entity_id=unique_id.make_unique_id(),
                )
                existing_entry = uw.repo.get_by_id_optional(entry.entity_id)
                if existing_entry and not existing_entry.discarded:
                    raise errors.IntegrityError(
                        f"An entry with entry_id '{entry.entity_id}' already exists."
                    )
                uw.entries.save(entry)
                created_db_entries.append(entry)

                if command.chunk_size > 0 and i % command.chunk_size == 0:
                    uw.commit()
            uw.commit()

        return created_db_entries


class ImportingEntries(BasingEntry, CommandHandler[commands.ImportEntries]):
    def execute(self, command: commands.ImportEntries):
        """
        Import entries to DB and INDEX (if present and resource is active).

        Raises
        ------
        RuntimeError
            If the resource.entry_json_schema fails to compile.
        KarpError
            - If an entry fails to be validated against the json schema.
            - If the DB interaction fails.

        Returns
        -------
        List
            List of the id's of the created entries.
        """

        if not isinstance(command.resource_id, str):
            raise ValueError(
                f"'resource_id' must be of type 'str', were '{type(command.resource_id)}'"
            )
        with self.resource_uow:
            resource = self.resource_uow.resources.by_resource_id(command.resource_id)

        if not resource:
            raise errors.ResourceNotFound(command.resource_id)

        entry_schema = EntrySchema(resource.entry_json_schema)

        created_db_entries = []
        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            resource.entry_repository_id
        ) as uw:
            for i, entry_raw in enumerate(command.entries):
                entry_schema.validate_entry(entry_raw["entry"])

                entry = resource.create_entry_from_dict(
                    entry_raw["entry"],
                    user=entry_raw.get("user") or command.user,
                    message=entry_raw.get("message") or command.message,
                    entity_id=entry_raw.get("entity_id") or unique_id.make_unique_id(),
                    timestamp=entry_raw.get("last_modified"),
                )
                # if entity_id := entry_raw.get("entity_id"):
                #     existing_entry = uw.repo.get_by_id_optional(entity_id)
                #     if existing_entry and not existing_entry.discarded:
                #         raise errors.IntegrityError(
                #             f"An entry with entry_id '{entry.entity_id}' already exists."
                #         )

                #             if entity_id != existing_entry.entity_id:
                #                 raise integrity_error
                #         else:
                #             raise integrity_error
                uw.entries.save(entry)
                created_db_entries.append(entry)

                if command.chunk_size > 0 and i % command.chunk_size == 0:
                    uw.commit()
            uw.commit()

        return created_db_entries


class DeletingEntry(BasingEntry, CommandHandler[commands.DeleteEntry]):
    def execute(self, command: commands.DeleteEntry):
        with self.resource_uow:
            resource = self.resource_uow.repo.by_resource_id(command.resource_id)

        with self.entry_repo_uow, self.entry_repo_uow.repo.get_by_id(
            resource.entry_repository_id
        ) as uw:
            entry = uw.repo.by_id(command.entity_id)

            entry.discard(
                user=command.user,
                message=command.message or "discarded",
                timestamp=command.timestamp,
            )
            uw.repo.save(entry)
            uw.commit()



from .entries import GetEntryDiff, GetEntryHistory, GetHistory, GetHistoryDto, EntryDiffDto, EntryDto, EntryViews, EntryDiffRequest, EntryHistoryRequest, HistoryDto
from .entry_repos import ListEntryRepos, EntryRepoDto, ReadOnlyEntryRepoRepositry
from .resources import GetPublishedResources, ResourceDto, GetResources, GetEntryRepositoryId, ReadOnlyResourceRepository
from .network import GetReferencedEntries, ReferenceDto



from karp.foundation import value_objects
from karp.lex.domain.entities.morphology import Morphology
from karp.lex.domain.entities.resource import Resource

from .factories import random_resource


def test_morphology():
    resource = random_resource()
    entry = {"baseform": "Appalacherna",
             "paradigm": "pm_plg_alperna", "pos": "pm"}

    # morphology = Morphology.create_resource(
    #     resource_type='SaldoMorphology',
    #     resource_config={},
    #     # entity_id=value_objects.make_unique_id(),
    # )
    # morphological_entry = {"entry_id": "pm_plg_alperna"}

    # needed_field_names = morphology.inflection_fields()

    # needed_fields = {
    #     needed_field_name: resource.get_field(needed_field_name, entry)
    #     for needed_field_name in needed_field_names
    # }
    # inflection_table = morphology.get_inflection_table(**needed_fields)
    # assert inflection_table == [
    #     {"form": "Appalacherna", "msd": "nom"},
    #     {"form": "Appalachernas", "msd": "gen"},
    # ]
